# -*- coding: utf-8 -*-

"""
This generate diagram of the (raw)ios and formats


Author: Julia Sprenger
"""

import pygraphviz
import neo

# from datetime import datetime
#
# import numpy as np
# import quantities as pq
# from matplotlib import pyplot
# from matplotlib.patches import Rectangle, ArrowStyle, FancyArrowPatch
# from matplotlib.font_manager import FontProperties
#
# from neo.test.generate_datasets import fake_neo
#
# line_heigth = .22
# fontsize = 10.5
# left_text_shift = .1
# dpi = 100


default_style = {"shape": "rectangle", "color": "black", "fontcolor": "black"}
IO_style = default_style.copy()
IO_style["fontsize"] = "30"
IO_style["penwidth"] = 6

styles = {
    "IO": {"ro": IO_style.copy(), "rw": IO_style.copy(), "raw": IO_style.copy()},
    "main": default_style.copy(),
    "ext": default_style.copy(),
}

styles["IO"]["ro"]["color"] = "#20B2AA "
styles["IO"]["rw"]["color"] = "#4169E1 "
styles["IO"]["raw"]["color"] = "#008080 "
styles["ext"]["shape"] = "circle"
styles["ext"]["fillcolor"] = "red"
styles["ext"]["style"] = "filled"


# styles['ext']['fixedsize'] = 'True'


def generate_diagram(filename, plot_extensions=False):
    dia = pygraphviz.AGraph(strict=False, splines="true")
    G = dia
    G.node_attr["fontname"] = "Arial"
    # G.node_attr['shape'] = 'circle'
    # G.node_attr['fixedsize'] = 'true'
    # G.node_attr['sep'] = '-100'
    G.node_attr["fixedsize"] = "False"
    # G.graph_attr['overlap'] = 'False'
    G.graph_attr["packMode"] = "clust"
    # G.graph_attr['levelsgap'] = -500
    G.node_attr["fontsize"] = "20"
    G.edge_attr["minlen"] = "0"
    # G.node_attr['style'] = 'filled'
    # G.graph_attr['outputorder'] = 'edgesfirst'
    # G.graph_attr['splines'] = "compound"
    G.graph_attr["label"] = "NEO {}".format(neo.__version__)
    G.graph_attr["ratio"] = "1.0"
    # G.edge_attr['color'] = '#1100FF'

    G.edge_attr["style"] = "setlinewidth(4)"

    dia.add_node("NEO", shape="circle", fontsize=50)

    for io in neo.io.iolist:
        io_name = io.name
        rawio = False
        if issubclass(io, neo.io.basefromrawio.BaseFromRaw):
            rawio = True
            if io_name == "BaseIO":
                io_name = io.__name__.rstrip("RawIO")
        if io_name is None:
            io_name = io.__name__.rstrip("IO")
        if "example" in io_name:
            continue

        if io.is_writable and io.is_readable:
            mode = "rw"
        elif io.is_readable:
            mode = "ro"
        if rawio:
            mode = "raw"

        dia.add_node(io_name, **styles["IO"][mode])
        dia.add_edge("NEO", io_name)

        if plot_extensions:
            for ext in io.extensions:
                dia.add_node(ext, **styles["ext"])
                dia.add_edge(io_name, ext, minlen=0)

    dia.layout(prog="fdp")  # neato, dot, twopi, circo, fdp, nop, wc, acyclic, gvpr, gvcolor,
    # ccomps, sccmap, tred, sfdp.
    for ext in ["png", "svg", "eps"]:
        dia.draw("{}.{}".format(filename, ext))


if __name__ == "__main__":
    generate_diagram("IODiagram", plot_extensions=False)
    generate_diagram("IODiagram_ext", plot_extensions=True)
    # pyplot.show()



"""
This class reads .set and .bin file data from the Axona acquisition system.

File format overview:
http://space-memory-navigation.org/DacqUSBFileFormats.pdf

In brief:
 data.set - setup file containing all hardware setups related to the trial
 data.bin - raw data file

There are many other data formats from Axona, which we do not consider (yet).
These are derivative from the raw continuous data (.bin) and could in principle
be extracted from it (see file format overview for details).

Author: Steffen Buergers, Julia Sprenger

"""

import datetime
import pathlib
import re
import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class AxonaRawIO(BaseRawIO):
    """
    Class for reading raw, continuous data from the Axona dacqUSB system

    Parameters
    ----------
    filename: str
        The name of the *.bin file containing the data

    Notes
    -----

    http://space-memory-navigation.org/DacqUSBFileFormats.pdf

    The raw data is saved in .bin binary files with an accompanying .set
    file about the recording setup (see the above manual for details).

    Examples
    --------

        >>> import neo.rawio
        >>> r = neo.rawio.AxonaRawIO(filename=os.path.join(dir_name, base_filename))
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                                 seg_index=0,
                                                 i_start=0,
                                                 i_stop=1024,
                                                 channel_names=channel_names)
        >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk,
                                                             dtype='float64',
                                                             channel_indexes=[0, 3, 6])

    """

    extensions = ["bin", "set"] + [str(i) for i in range(1, 33)]  # Never used?
    rawmode = "multi-file"

    # In the .bin file, channels are arranged in a strange order.
    # This list takes a channel index as input and returns the actual
    # offset for the channel in the memory map (self._raw_signals).
    channel_memory_offset = [
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        39,
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        40,
        41,
        42,
        43,
        44,
        45,
        46,
        47,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        56,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
    ]

    def __init__(self, filename):
        BaseRawIO.__init__(self)

        # Accepting filename with arbitrary suffix as input
        self.filename = pathlib.Path(filename).with_suffix("")
        self.set_file = self.filename.with_suffix(".set")
        self.bin_file = None
        self.tetrode_files = []

        # set file is mandatory for any recording
        if not self.set_file.exists():
            raise ValueError(
                f'Could not locate ".set" file. ' f'{self.filename.with_suffix(".set")} does not ' f"exist."
            )

        # detecting available files
        if self.filename.with_suffix(".bin").exists():
            self.bin_file = self.filename.with_suffix(".bin")

        for i in range(1, 33):
            unit_file = self.filename.with_suffix(f".{i}")
            if unit_file.exists():
                self.tetrode_files.append(unit_file)
            else:
                break

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        """
        Read important information from .set header file, create memory map
        to raw data (.bin file) and prepare header dictionary in neo format.
        """

        unit_dtype = np.dtype([("spiketimes", ">i4"), ("samples", "int8", (50,))])

        # Utility collection of file parameters (general info and header data)
        params = {
            "bin": {
                "filename": self.bin_file,
                "bytes_packet": 432,
                "bytes_data": 384,
                "bytes_head": 32,
                "bytes_tail": 16,
                "data_type": "int16",
                "header_size": 0,
                # bin files don't contain a file header
                "header_encoding": None,
            },
            "set": {"filename": self.set_file, "header_encoding": "cp1252"},
            "unit": {"data_type": unit_dtype, "tetrode_ids": [], "header_encoding": "cp1252"},
        }
        self.file_parameters = params

        # SCAN SET FILE
        set_dict = self.get_header_parameters(self.set_file, "set")
        params["set"]["file_header"] = set_dict
        params["set"]["sampling_rate"] = int(set_dict["rawRate"])

        # SCAN BIN FILE
        signal_streams = []
        signal_channels = []
        if self.bin_file:
            bin_dict = self.file_parameters["bin"]
            # add derived parameters from bin file
            bin_dict["num_channels"] = len(self.get_active_tetrode()) * 4
            num_tot_packets = int(self.bin_file.stat().st_size / bin_dict["bytes_packet"])
            bin_dict["num_total_packets"] = num_tot_packets
            bin_dict["num_total_samples"] = num_tot_packets * 3

            # Create np.memmap to .bin file
            self._raw_signals = np.memmap(
                self.bin_file,
                dtype=self.file_parameters["bin"]["data_type"],
                mode="r",
                offset=self.file_parameters["bin"]["header_size"],
            )

            signal_streams = self._get_signal_streams_header()
            signal_channels = self._get_signal_chan_header()

        # SCAN TETRODE FILES
        # In this IO one tetrode corresponds to one unit as spikes are not
        # sorted yet.
        self._raw_spikes = []
        spike_channels = []
        if self.tetrode_files:

            for i, tetrode_file in enumerate(self.tetrode_files):
                # collecting tetrode specific parameters and dtype conversions
                tdict = self.get_header_parameters(tetrode_file, "unit")
                tdict["filename"] = tetrode_file
                tdict["num_chans"] = int(tdict["num_chans"])
                tdict["num_spikes"] = int(tdict["num_spikes"])
                tdict["header_size"] = len(self.get_header_bstring(tetrode_file))

                # memory mapping spiking data
                spikes = np.memmap(
                    tetrode_file,
                    dtype=self.file_parameters["unit"]["data_type"],
                    mode="r",
                    offset=tdict["header_size"],
                    shape=(tdict["num_spikes"], 4),
                )
                self._raw_spikes.append(spikes)

                unit_name = f"tetrode {i + 1}"
                unit_id = f"{i + 1}"
                wf_units = "dimensionless"
                wf_gain = 1
                wf_offset = 0.0
                # left sweep information is only stored in set file
                wf_left_sweep = int(self.file_parameters["set"]["file_header"]["pretrigSamps"])

                # Extract waveform sample rate
                # 1st priority source: Spike2msMode (0 -> 48kHz; 1 -> 24kHz)
                # 2nd priority source: tetrode sample rate
                spikemode_to_sr = {0: 48000, 1: 24000}  # spikemode->rate in Hz
                sm = self.file_parameters["set"]["file_header"].get("Spike2msMode", -1)
                wf_sampling_rate = spikemode_to_sr.get(int(sm), None)
                if wf_sampling_rate is None:
                    wf_sampling_rate = self._to_hz(tdict["sample_rate"], dtype=float)

                spike_channels.append(
                    (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                )

                self.file_parameters["unit"]["tetrode_ids"].append(i + 1)
                self.file_parameters["unit"][i + 1] = tdict

            # propagate common tetrode parameters to global unit level
            units_dict = self.file_parameters["unit"]
            ids = units_dict["tetrode_ids"]
            copied_keys = []
            if ids:
                for key, value in units_dict[ids[0]].items():
                    # copy key-value pair if present across all tetrodes
                    if all([key in units_dict[t] for t in ids]) and all([units_dict[t][key] == value for t in ids]):
                        self.file_parameters["unit"][key] = value
                        copied_keys.append(key)

                # remove key from individual tetrode parameters
                for key in copied_keys:
                    for t in ids:
                        self.file_parameters["unit"][t].pop(key)

        # Create RawIO header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = np.array(signal_streams, dtype=_signal_stream_dtype)
        self.header["signal_channels"] = np.array(signal_channels, dtype=_signal_channel_dtype)
        self.header["spike_channels"] = np.array(spike_channels, dtype=_spike_channel_dtype)
        self.header["event_channels"] = np.array([], dtype=_event_channel_dtype)

        # Annotations
        self._generate_minimal_annotations()

        # Adding custom annotations
        bl_ann = self.raw_annotations["blocks"][0]
        seg_ann = bl_ann["segments"][0]
        seg_ann["rec_datetime"] = self.read_datetime()
        if len(seg_ann["signals"]):
            seg_ann["signals"][0]["__array_annotations__"]["tetrode_id"] = [
                tetr for tetr in self.get_active_tetrode() for _ in range(4)
            ]

        if len(seg_ann["spikes"]):
            # adding segment annotations
            seg_keys = ["experimenter", "comments", "sw_version"]
            for seg_key in seg_keys:
                if seg_key in self.file_parameters["unit"]:
                    seg_ann[seg_key] = self.file_parameters["unit"][seg_key]

    def _get_signal_streams_header(self):
        # create signals stream information (we always expect a single stream)
        return np.array([("stream 0", "0")], dtype=_signal_stream_dtype)

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = 0.0

        if "num_total_packets" in self.file_parameters["bin"]:
            sr = self.file_parameters["set"]["sampling_rate"]
            t_stop = self.file_parameters["bin"]["num_total_samples"] / sr

        if self.file_parameters["unit"]["tetrode_ids"]:
            # get tetrode recording durations in seconds
            if "duration" not in self.file_parameters["unit"]:
                raise ValueError("Can not determine common tetrode recording" "duration.")

            tetrode_duration = float(self.file_parameters["unit"]["duration"])
            t_stop = max(t_stop, tetrode_duration)

        return t_stop

    def _get_signal_size(self, block_index, seg_index, channel_indexes=None):
        if "num_total_packets" in self.file_parameters["bin"]:
            return self.file_parameters["bin"]["num_total_samples"]
        else:
            return 0

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        """
        Return raw (continuous) signals as 2d numpy array (time x chan).
        Note that block_index and seg_index are always 1 (regardless of input).

        Raw data is in a single vector np.memmap with the following structure:

        Each byte packet (432 bytes) has header (32 bytes), footer (16 bytes)
        and three samples of 2 bytes each for 64 channels (384 bytes), which
        are jumbled up in a strange order. Each channel is remapped to a
        certain position (see get_channel_offset), and a channel's samples are
        allocated as follows (example for channel 7):

        sample 1: 32b (head) + 2*38b (remappedID) and 2*38b + 1b (2nd byte)
        sample 2: 32b (head) + 128 (all chan. 1st entry) + 2*38b and ...
        sample 3: 32b (head) + 128*2 (all channels 1st and 2nd entry) + ...
        """

        bin_dict = self.file_parameters["bin"]

        # Set default values
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = bin_dict["num_total_samples"]
        if channel_indexes is None:
            channel_indexes = [i for i in range(bin_dict["num_channels"])]
        elif isinstance(channel_indexes, slice):
            channel_indexes_all = [i for i in range(bin_dict["num_channels"])]
            channel_indexes = channel_indexes_all[channel_indexes]

        num_samples = i_stop - i_start

        # Create base index vector for _raw_signals for time period of interest
        num_packets_oi = (num_samples + 2) // 3
        offset = i_start // 3 * (bin_dict["bytes_packet"] // 2)
        rem = i_start % 3

        raw_samples = np.arange(num_packets_oi + 1, dtype=np.uint32)
        sample1 = raw_samples * (bin_dict["bytes_packet"] // 2) + bin_dict["bytes_head"] // 2 + offset
        sample2 = sample1 + 64
        sample3 = sample2 + 64

        sig_ids = np.empty((sample1.size + sample2.size + sample3.size,), dtype=sample1.dtype)
        sig_ids[0::3] = sample1
        sig_ids[1::3] = sample2
        sig_ids[2::3] = sample3
        sig_ids = sig_ids[rem : (rem + num_samples)]

        # Read one channel at a time
        raw_signals = np.ndarray(shape=(num_samples, len(channel_indexes)), dtype=bin_dict["data_type"])

        for i, ch_idx in enumerate(channel_indexes):
            chan_offset = self.channel_memory_offset[ch_idx]
            raw_signals[:, i] = self._raw_signals[sig_ids + chan_offset]

        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):
        tetrode_id = unit_index

        raw_spikes = self._raw_spikes[tetrode_id]
        nb_tetrode_spikes = raw_spikes.shape[0]
        # also take into account last, potentially incomplete set of spikes
        nb_unit_spikes = int(np.ceil(nb_tetrode_spikes))
        return nb_unit_spikes

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        if block_index != 0:
            raise ValueError("`block_index must be 0")
        if seg_index != 0:
            raise ValueError("`seg_index` must be 0")

        tetrode_id = unit_index
        raw_spikes = self._raw_spikes[tetrode_id]

        # spike times are repeated for each contact -> use only first contact
        unit_spikes = raw_spikes["spiketimes"][:, 0]

        # slice spike times only if needed
        if t_start is None and t_stop is None:
            return unit_spikes

        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)

        mask = self._get_temporal_mask(t_start, t_stop, tetrode_id)

        return unit_spikes[mask]

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self._to_hz(self.file_parameters["unit"]["timebase"], dtype=int)
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        if block_index != 0:
            raise ValueError("`block_index must be 0")
        if seg_index != 0:
            raise ValueError("`seg_index` must be 0")

        tetrode_id = unit_index
        waveforms = self._raw_spikes[tetrode_id]["samples"]

        # slice timestamps / waveforms only when necessary
        if t_start is None and t_stop is None:
            return waveforms

        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)

        mask = self._get_temporal_mask(t_start, t_stop, tetrode_id)
        waveforms = waveforms[mask]

        return waveforms

    def get_header_bstring(self, file):
        """
        Scan file for the occurrence of 'data_start' and return the header
        as byte string

        Parameters
        ----------
        file (str or path): file to be loaded

        Returns
        -------
        str: header byte content
        """

        header = b""
        with open(file, "rb") as f:
            for bin_line in f:
                if b"data_start" in bin_line:
                    header += b"data_start"
                    break
                else:
                    header += bin_line
        return header

    # ------------------ HELPER METHODS --------------------
    # This is largely based on code by Geoff Barrett from the Hussaini lab:
    # https://github.com/GeoffBarrett/BinConverter
    # Adapted or modified by Steffen Buergers, Julia Sprenger

    def _get_temporal_mask(self, t_start, t_stop, tetrode_id):
        # Convenience function for creating a temporal mask given
        # start time (t_start) and stop time (t_stop)
        # Used by _get_spike_raw_waveforms and _get_spike_timestamps

        # spike times are repeated for each contact -> use only first contact
        raw_spikes = self._raw_spikes[tetrode_id]
        unit_spikes = raw_spikes["spiketimes"][:, 0]

        # convert t_start and t_stop to sampling frequency
        # Note: this assumes no time offset!
        unit_params = self.file_parameters["unit"]
        lim0 = t_start * self._to_hz(unit_params["timebase"], dtype=int)
        lim1 = t_stop * self._to_hz(unit_params["timebase"], dtype=int)

        mask = (unit_spikes >= lim0) & (unit_spikes <= lim1)

        return mask

    def get_header_parameters(self, file, file_type):
        """
        Extract header parameters as dictionary keys and
        following phrases as values (strings).

        Parameters
        ----------
        file (str or path): file to be loaded
        file_type (str): type of file to be loaded ('set' or 'unit')

        Returns
        -------
        header (dict): dictionary with keys being the parameters that
                       were found & values being strings of the data.

        EXAMPLE
        self.get_header_parameters('file.set', 'set')
        """

        params = {}
        encoding = self.file_parameters[file_type]["header_encoding"]
        header_string = self.get_header_bstring(file).decode(encoding)

        # omit the last line as this contains only `data_start` key
        for line in header_string.splitlines()[:-1]:
            key, value = line.split(" ", 1)
            params[key] = value
        return params

    def get_active_tetrode(self):
        """
        Returns the ID numbers of the active tetrodes as a list.
        E.g.: [1,2,3,4] for a recording with 4 tetrodes (16 channels).
        """
        active_tetrodes = []

        for key, status in self.file_parameters["set"]["file_header"].items():
            # The pattern to look for is collectMask_X Y,
            # where X is the tetrode number, and Y is 1 or 0
            if key.startswith("collectMask_"):
                if int(status):
                    tetrode_id = int(key.strip("collectMask_"))
                    active_tetrodes.append(tetrode_id)
        return active_tetrodes

    def _get_channel_from_tetrode(self, tetrode):
        """
        This function will take the tetrode number and return the Axona
        channel numbers, i.e. Tetrode 1 = Ch0-Ch3, Tetrode 2 = Ch4-Ch7, etc.
        """
        return np.arange(0, 4) + 4 * (int(tetrode) - 1)

    def read_datetime(self):
        """
        Creates datetime object (y, m, d, h, m, s) from .set file header
        """

        date_str = self.file_parameters["set"]["file_header"]["trial_date"]
        time_str = self.file_parameters["set"]["file_header"]["trial_time"]

        # extract core date string
        date_str = re.findall(r"\d+\s\w+\s\d{4}$", date_str)[0]

        return datetime.datetime.strptime(date_str + ", " + time_str, "%d %b %Y, %H:%M:%S")

    def _get_channel_gain(self, bytes_per_sample=2):
        """
        This is actually not the gain_ch value from the .set file, but the
        conversion factor from raw data to uV.

        Formula for conversion to uV:

        1000 * adc_fullscale_mv / (gain_ch * max-value), with
        max_value = 2**(8 * bytes_per_sample - 1)

        Adapted from
        https://github.com/CINPLA/pyxona/blob/stable/pyxona/core.py
        """
        gain_list = []

        adc_fm = int(self.file_parameters["set"]["file_header"]["ADC_fullscale_mv"])

        for key, value in self.file_parameters["set"]["file_header"].items():
            if key.startswith("gain_ch"):
                gain_list.append(np.float32(value))

        max_value = 2 ** (8 * bytes_per_sample - 1)
        gain_list = [1000 * adc_fm / (gain * max_value) for gain in gain_list]

        return gain_list

    def _get_signal_chan_header(self):
        """
        Returns a 1 dimensional np.array of tuples with one entry per channel
        that recorded data. Each tuple contains the following information:

        channel name (1a, 1b, 1c, 1d, 2a, 2b, ...; num=tetrode, letter=elec),
        channel id (1, 2, 3, 4, 5, ... N),
        sampling rate,
        data type (int16),
        unit (uV),
        gain,
        offset,
        stream id
        """
        active_tetrode_set = self.get_active_tetrode()
        num_active_tetrode = len(active_tetrode_set)

        elec_per_tetrode = 4
        letters = ["a", "b", "c", "d"]
        dtype = self.file_parameters["bin"]["data_type"]
        units = "uV"
        gain_list = self._get_channel_gain()
        offset = 0  # What is the offset?

        sig_channels = []
        for itetr in range(num_active_tetrode):

            for ielec in range(elec_per_tetrode):
                cntr = (itetr * elec_per_tetrode) + ielec
                ch_name = f"{itetr + 1}{letters[ielec]}"
                chan_id = str(cntr)
                gain = gain_list[cntr]
                stream_id = "0"
                # the sampling rate information is stored in the set header
                # and not in the bin file
                sr = self.file_parameters["set"]["sampling_rate"]
                sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))

        return np.array(sig_channels, dtype=_signal_channel_dtype)

    def _to_hz(self, param, dtype=float):
        return dtype(param.replace(" hz", ""))



"""
Generate datasets for testing
"""

from datetime import datetime
import random
import string
import numpy as np
from numpy.random import rand
import quantities as pq

from neo.core import (
    AnalogSignal,
    Block,
    Epoch,
    Event,
    IrregularlySampledSignal,
    Group,
    Segment,
    SpikeTrain,
    ImageSequence,
    ChannelView,
    CircularRegionOfInterest,
)

TEST_ANNOTATIONS = [1, 0, 1.5, "this is a test", datetime.fromtimestamp(424242424), None]


def random_string(length=10):
    return "".join(random.choice(string.ascii_letters) for i in range(length))


def random_datetime(min_year=1990, max_year=datetime.now().year):
    start = datetime(min_year, 1, 1, 0, 0, 0)
    end = datetime(max_year, 12, 31, 23, 59, 59)
    return start + (end - start) * random.random()


def random_annotations(n=1):
    annotation_generators = (random.random, random_datetime, random_string, lambda: None)
    annotations = {}
    for i in range(n):
        var_name = random_string(6)
        annotation_generator = random.choice(annotation_generators)
        annotations[var_name] = annotation_generator()
    return annotations


def random_signal(name=None, **annotations):
    n_channels = random.randint(1, 7)
    sig_length = random.randint(20, 200)
    if len(annotations) == 0:
        annotations = random_annotations(5)
    obj = AnalogSignal(
        np.random.uniform(size=(sig_length, n_channels)),
        units=random.choice(("mV", "nA")),
        t_start=random.uniform(0, 10) * pq.ms,
        sampling_rate=random.uniform(0.1, 10) * pq.kHz,
        name=name or random_string(),
        file_origin=random_string(),
        description=random_string(100),
        array_annotations=None,  # todo
        **annotations,
    )
    return obj


def random_irreg_signal(name=None, **annotations):
    n_channels = random.randint(1, 7)
    sig_length = random.randint(20, 200)
    if len(annotations) == 0:
        annotations = random_annotations(5)
    mean_firing_rate = np.random.uniform(0.1, 10) * pq.kHz
    times = np.cumsum(np.random.uniform(1.0 / mean_firing_rate, size=(sig_length,))) * pq.ms
    obj = IrregularlySampledSignal(
        times,
        np.random.uniform(size=(sig_length, n_channels)),
        units=random.choice(("mV", "nA")),
        name=name or random_string(),
        file_origin=random_string(),
        description=random_string(100),
        array_annotations=None,  # todo
        **annotations,
    )
    return obj


def random_event(name=None, **annotations):
    size = random.randint(1, 7)
    times = np.cumsum(np.random.uniform(5, 10, size=size))
    labels = [random_string() for i in range(size)]
    if len(annotations) == 0:
        annotations = random_annotations(3)
    obj = Event(
        times=times,
        labels=labels,
        units="ms",
        name=name or random_string(),
        array_annotations=None,  # todo
        **annotations,
    )
    return obj


def random_epoch():
    size = random.randint(1, 7)
    times = np.cumsum(np.random.uniform(5, 10, size=size))
    durations = np.random.uniform(1, 3, size=size)
    labels = [random_string() for i in range(size)]
    obj = Epoch(
        times=times,
        durations=durations,
        labels=labels,
        units="ms",
        name=random_string(),
        array_annotations=None,  # todo
        **random_annotations(3),
    )
    return obj


def random_spiketrain(name=None, **annotations):
    size = random.randint(1, 50)
    times = np.cumsum(np.random.uniform(0.5, 10, size=size))
    if len(annotations) == 0:
        annotations = random_annotations(3)
    # todo: waveforms
    obj = SpikeTrain(
        times=times,
        t_stop=times[-1] + random.uniform(0, 5),
        units="ms",
        name=name or random_string(),
        array_annotations=None,  # todo
        **annotations,
    )
    return obj


def random_image_sequence(name=None, **annotations):
    pixels_i = random.randint(2, 7)
    pixels_j = random.randint(2, 7)
    seq_length = random.randint(20, 200)
    if len(annotations) == 0:
        annotations = random_annotations(5)
    obj = ImageSequence(
        np.random.uniform(size=(seq_length, pixels_i, pixels_j)),
        t_start=random.uniform(0, 10) * pq.ms,
        sampling_rate=random.uniform(0.1, 10) * pq.kHz,
        spatial_scale=random.uniform(0.1, 10) * pq.um,
        name=name or random_string(),
        file_origin=random_string(),
        description=random_string(100),
        array_annotations=None,  # todo
        **annotations,
    )
    return obj


def random_segment():
    seg = Segment(
        name=random_string(10),
        description=random_string(100),
        file_origin=random_string(20),
        file_datetime=random_datetime(),
        rec_datetime=random_datetime(),
        **random_annotations(4),
    )
    n_sigs = random.randint(0, 5)
    for i in range(n_sigs):
        seg.analogsignals.append(random_signal())
    n_irrsigs = random.randint(0, 5)
    for i in range(n_irrsigs):
        seg.irregularlysampledsignals.append(random_irreg_signal())
    n_events = random.randint(0, 3)
    for i in range(n_events):
        seg.events.append(random_event())
    n_epochs = random.randint(0, 3)
    for i in range(n_epochs):
        seg.epochs.append(random_epoch())
    n_spiketrains = random.randint(0, 20)
    for i in range(n_spiketrains):
        seg.spiketrains.append(random_spiketrain())
    n_imgs = random.randint(0, 5)
    for i in range(n_imgs):
        seg.imagesequences.append(random_image_sequence())
    # todo: add some and ROI objects

    return seg


def random_group(candidates):
    if len(candidates) == 0:
        return None
    elif len(candidates) == 1:
        objects = candidates
    else:
        k = random.randint(1, len(candidates))
        objects = random.sample(candidates, k)
    obj = Group(objects=objects, name=random_string(), **random_annotations(5))
    return obj


def random_channelview(signal):
    n_channels = signal.shape[1]
    if n_channels > 2:
        view_size = np.random.randint(1, n_channels - 1)
        index = np.random.choice(np.arange(signal.shape[1]), view_size, replace=False)
        obj = ChannelView(signal, index, name=random_string(), **random_annotations(3))
        return obj
    else:
        return None


def random_roi(imgseq):
    x = np.random.randint(imgseq.shape[1])
    y = np.random.randint(imgseq.shape[2])
    radius = np.random.uniform() * imgseq.shape[1]
    obj = CircularRegionOfInterest(imgseq, x, y, radius, name=random_string(), **random_annotations(3))
    return obj


def random_block():
    block = Block(
        name=random_string(10),
        description=random_string(100),
        file_origin=random_string(20),
        file_datetime=random_datetime(),
        rec_datetime=random_datetime(),
        **random_annotations(6),
    )
    n_seg = random.randint(0, 5)
    for i in range(n_seg):
        seg = random_segment()
        block.segments.append(seg)
    children = list(block.data_children_recur)
    views = []
    for child in children:
        if isinstance(child, (AnalogSignal, IrregularlySampledSignal)):
            PROB_SIGNAL_HAS_VIEW = 0.5
            if np.random.random_sample() < PROB_SIGNAL_HAS_VIEW:
                chv = random_channelview(child)
                if chv:
                    views.append(chv)
        elif isinstance(child, ImageSequence):
            PROB_IMGSEQ_HAS_ROI = 0.5
            if np.random.random_sample() < PROB_IMGSEQ_HAS_ROI:
                roi = random_roi(child)
                if roi:
                    views.append(roi)
    children.extend(views)
    n_groups = random.randint(0, 5)
    for i in range(n_groups):
        group = random_group(children)
        if group:
            block.groups.append(group)
            children.append(group)  # this can give us nested groups
    return block


def simple_block():
    block = Block(name="test block", species="rat", brain_region="cortex")
    block.segments = [
        Segment(name="test segment #1", cell_type="spiny stellate"),
        Segment(name="test segment #2", cell_type="pyramidal", thing="amajig"),
    ]
    block.segments[0].analogsignals.extend(
        (
            random_signal(name="signal #1 in segment #1", thing="wotsit"),
            random_signal(name="signal #2 in segment #1", thing="frooble"),
        )
    )
    block.segments[1].analogsignals.extend((random_signal(name="signal #1 in segment #2", thing="amajig"),))
    block.segments[1].irregularlysampledsignals.extend(
        (random_irreg_signal(name="signal #1 in segment #2", thing="amajig"),)
    )
    block.segments[0].events.extend((random_event(name="event array #1 in segment #1", thing="frooble"),))
    block.segments[1].events.extend((random_event(name="event array #1 in segment #2", thing="wotsit"),))
    block.segments[0].spiketrains.extend(
        (
            random_spiketrain(name="spiketrain #1 in segment #1", thing="frooble"),
            random_spiketrain(name="spiketrain #2 in segment #1", thing="wotsit"),
        )
    )
    return block


def generate_one_simple_block(block_name="block_0", nb_segment=3, supported_objects=[], **kws):
    if supported_objects and Block not in supported_objects:
        raise ValueError("Block must be in supported_objects")
    bl = Block()  # name = block_name)

    objects = supported_objects
    if Segment in objects:
        for s in range(nb_segment):
            seg = generate_one_simple_segment(seg_name="seg" + str(s), supported_objects=objects, **kws)
            bl.segments.append(seg)

    bl.check_relationships()
    return bl


def generate_one_simple_segment(
    seg_name="segment 0",
    supported_objects=[],
    nb_analogsignal=4,
    t_start=0.0 * pq.s,
    sampling_rate=10 * pq.kHz,
    duration=6.0 * pq.s,
    nb_spiketrain=6,
    spikerate_range=[0.5 * pq.Hz, 12 * pq.Hz],
    event_types={
        "stim": ["a", "b", "c", "d"],
        "enter_zone": ["one", "two"],
        "color": ["black", "yellow", "green"],
    },
    event_size_range=[5, 20],
    epoch_types={"animal state": ["Sleep", "Freeze", "Escape"], "light": ["dark", "lighted"]},
    epoch_duration_range=[0.5, 3.0],
    # this should be multiplied by pq.s, no?
    array_annotations={"valid": np.array([True, False]), "number": np.array(range(5))},
):
    if supported_objects and Segment not in supported_objects:
        raise ValueError("Segment must be in supported_objects")
    seg = Segment(name=seg_name)
    if AnalogSignal in supported_objects:
        for a in range(nb_analogsignal):
            anasig = AnalogSignal(
                rand(int((sampling_rate * duration).simplified)),
                sampling_rate=sampling_rate,
                t_start=t_start,
                units=pq.mV,
                channel_index=a,
                name=f"sig {a} for segment {seg.name}",
            )
            seg.analogsignals.append(anasig)

    if SpikeTrain in supported_objects:
        for s in range(nb_spiketrain):
            spikerate = rand() * np.diff(spikerate_range)[0]
            spikerate += spikerate_range[0].item()
            spikes = rand(int((spikerate * duration).simplified))
            spikes.sort()  # spikes are supposed to be an ascending sequence
            sptr = SpikeTrain(spikes * duration, t_start=t_start, t_stop=t_start + duration)
            sptr.annotations["channel_index"] = s
            # Randomly generate array_annotations from given options
            arr_ann = {
                key: value[(rand(len(spikes)) * len(value)).astype("i")] for (key, value) in array_annotations.items()
            }
            sptr.array_annotate(**arr_ann)
            seg.spiketrains.append(sptr)

    if Event in supported_objects:
        for name, labels in event_types.items():
            evt_size = rand() * np.diff(event_size_range)[0]
            evt_size += event_size_range[0]
            evt_size = int(evt_size)
            labels = np.array(labels, dtype="U")
            labels = labels[(rand(evt_size) * len(labels)).astype("i")]
            evt = Event(times=rand(evt_size) * duration, labels=labels)
            # Randomly generate array_annotations from given options
            arr_ann = {
                key: value[(rand(evt_size) * len(value)).astype("i")] for (key, value) in array_annotations.items()
            }
            evt.array_annotate(**arr_ann)
            seg.events.append(evt)

    if Epoch in supported_objects:
        for name, labels in epoch_types.items():
            t = 0
            times = []
            durations = []
            while t < duration:
                times.append(t)
                dur = rand() * (epoch_duration_range[1] - epoch_duration_range[0])
                dur += epoch_duration_range[0]
                durations.append(dur)
                t = t + dur
            labels = np.array(labels, dtype="U")
            labels = labels[(rand(len(times)) * len(labels)).astype("i")]
            assert len(times) == len(durations)
            assert len(times) == len(labels)
            epc = Epoch(
                times=pq.Quantity(times, units=pq.s),
                durations=pq.Quantity(durations, units=pq.s),
                labels=labels,
            )
            assert epc.times.dtype == "float"
            # Randomly generate array_annotations from given options
            arr_ann = {
                key: value[(rand(len(times)) * len(value)).astype("i")] for (key, value) in array_annotations.items()
            }
            epc.array_annotate(**arr_ann)
            seg.epochs.append(epc)

    # TODO : Spike, Event

    seg.check_relationships()
    return seg


def generate_from_supported_objects(supported_objects):
    if not supported_objects:
        raise ValueError("No objects specified")
    objects = supported_objects
    if Block in supported_objects:
        higher = generate_one_simple_block(supported_objects=objects)
    elif Segment in objects:
        higher = generate_one_simple_segment(supported_objects=objects)
    else:
        # TODO
        return None

    higher.check_relationships()
    return higher



"""
Class for reading data from WinWCP, a software tool written by
John Dempster.

WinWCP is free:
http://spider.science.strath.ac.uk/sipbs/software.htm

Author: Samuel Garcia
"""

import struct

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
    _common_sig_characteristics,
)


class WinWcpRawIO(BaseRawIO):
    """
    Class for reading WinWCP data

    Parameters
    ----------
    filename: str, default: ''
        The *.wcp file to load

    """

    extensions = ["wcp"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        SECTORSIZE = 512

        # only one memmap for all segment to avoid
        # "error: [Errno 24] Too many open files"
        self._memmap = np.memmap(self.filename, dtype="uint8", mode="r")

        with open(self.filename, "rb") as fid:

            headertext = fid.read(1024)
            headertext = headertext.decode("ascii")
            header = {}
            for line in headertext.split("\r\n"):
                if "=" not in line:
                    continue
                # print '#' , line , '#'
                key, val = line.split("=")
                if key in [
                    "NC",
                    "NR",
                    "NBH",
                    "NBA",
                    "NBD",
                    "ADCMAX",
                    "NP",
                    "NZ",
                ]:
                    val = int(val)
                elif key in [
                    "AD",
                    "DT",
                ]:
                    val = val.replace(",", ".")
                    val = float(val)
                header[key] = val

            nb_segment = header["NR"]
            self._raw_signals = {}
            all_sampling_interval = []
            # loop for record number
            for seg_index in range(header["NR"]):
                offset = 1024 + seg_index * (SECTORSIZE * header["NBD"] + 1024)

                # read analysis zone
                analysisHeader = HeaderReader(fid, AnalysisDescription).read_f(offset=offset)

                # read data
                NP = (SECTORSIZE * header["NBD"]) // 2
                NP = NP - NP % header["NC"]
                NP = NP // header["NC"]
                NC = header["NC"]
                ind0 = offset + header["NBA"] * SECTORSIZE
                ind1 = ind0 + NP * NC * 2
                sigs = self._memmap[ind0:ind1].view("int16").reshape(NP, NC)
                self._raw_signals[seg_index] = sigs

                all_sampling_interval.append(analysisHeader["SamplingInterval"])

        # sampling interval can be slightly varying due to float precision
        # all_sampling_interval are not always unique
        self._sampling_rate = 1.0 / np.median(all_sampling_interval)

        signal_channels = []
        for c in range(header["NC"]):
            YG = float(header[f"YG{c}"].replace(",", "."))
            ADCMAX = header["ADCMAX"]
            VMax = analysisHeader["VMax"][c]

            name = header[f"YN{c}"]
            chan_id = header[f"YO{c}"]
            units = header[f"YU{c}"]
            gain = VMax / ADCMAX / YG
            offset = 0.0
            stream_id = "0"
            signal_channels.append((name, chan_id, self._sampling_rate, "int16", units, gain, offset, stream_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        characteristics = signal_channels[_common_sig_characteristics]
        unique_characteristics = np.unique(characteristics)
        signal_streams = []
        for i in range(unique_characteristics.size):
            mask = unique_characteristics[i] == characteristics
            signal_channels["stream_id"][mask] = str(i)
            signal_streams.append((f"stream {i}", str(i)))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals[seg_index].shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._raw_signals[seg_index].shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        (global_channel_indexes,) = np.nonzero(self.header["signal_channels"]["stream_id"] == stream_id)
        if channel_indexes is None:
            channel_indexes = slice(None)
        inds = global_channel_indexes[channel_indexes]
        raw_signals = self._raw_signals[seg_index][slice(i_start, i_stop), inds]
        return raw_signals


AnalysisDescription = [
    ("RecordStatus", "8s"),
    ("RecordType", "4s"),
    ("GroupNumber", "f"),
    ("TimeRecorded", "f"),
    ("SamplingInterval", "f"),
    ("VMax", "8f"),
]


class HeaderReader:
    def __init__(self, fid, description):
        self.fid = fid
        self.description = description

    def read_f(self, offset=0):
        self.fid.seek(offset)
        d = {}
        for key, fmt in self.description:
            val = struct.unpack(fmt, self.fid.read(struct.calcsize(fmt)))
            if len(val) == 1:
                val = val[0]
            else:
                val = list(val)
            d[key] = val
        return d



"""
PhyRawIO is a class to handle Phy spike sorting data.
Ported from:
https://github.com/SpikeInterface/spikeextractors/blob/
f20b1219eba9d3330d5d7cd7ce8d8924a255b8c2/spikeextractors/
extractors/phyextractors/phyextractors.py

Author: Regimantas Jurkus
"""

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

import numpy as np
from pathlib import Path
import re
import csv
import ast
import warnings


class PhyRawIO(BaseRawIO):
    """
    Class for reading Phy data.

    Parameters
    ----------
    dirname: str | Path, default: ''
        The dirname containing the phy data (.npy, .mat, .tsv. & .dat files)
    load_amplitudes: bool, default: False
        Whether to load the amplitude data
    load_pcs: bool, default: False
        Whether to load the pcs feature files (these are relatively big files)

    Examples
    --------
    >>> import neo.rawio
    >>> r = neo.rawio.PhyRawIO(dirname='/dir/to/phy/folder')
    >>> r.parse_header()
    >>> print(r)
    >>> spike_timestamp = r.get_spike_timestamps(block_index=0,
                                                 seg_index=0,
                                                 spike_channel_index=0,
                                                 t_start=None,
                                                 t_stop=None)
    >>> spike_times = r.rescale_spike_timestamp(spike_timestamp, dtype='float64')

    """

    # file formats used by phy
    extensions = ["npy", "mat", "tsv", "dat"]
    rawmode = "one-dir"

    def __init__(self, dirname="", load_amplitudes=False, load_pcs=False):
        BaseRawIO.__init__(self)
        self.dirname = dirname
        self.load_pcs = load_pcs
        self.load_amplitudes = load_amplitudes

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        phy_folder = Path(self.dirname)

        self._spike_times = np.load(phy_folder / "spike_times.npy")
        self._spike_templates = np.load(phy_folder / "spike_templates.npy")

        if (phy_folder / "spike_clusters.npy").is_file():
            self._spike_clusters = np.load(phy_folder / "spike_clusters.npy")
        else:
            self._spike_clusters = self._spike_templates

        self._amplitudes = None
        if self.load_amplitudes:
            if (phy_folder / "amplitudes.npy").is_file():
                self._amplitudes = np.squeeze(np.load(phy_folder / "amplitudes.npy"))
            else:
                warnings.warn('Amplitudes requested but "amplitudes.npy"' "not found in the data folder.")

        self._pc_features = None
        self._pc_feature_ind = None
        if self.load_pcs:
            if (phy_folder / "pc_features.npy").is_file() and (phy_folder / "pc_feature_ind.npy").is_file():
                self._pc_features = np.squeeze(np.load(phy_folder / "pc_features.npy"))
                self._pc_feature_ind = np.squeeze(np.load(phy_folder / "pc_feature_ind.npy"))
            else:
                warnings.warn(
                    'PCs requested but "pc_features.npy" and/or' '"pc_feature_ind.npy" not found in the data folder.'
                )

        # SEE: https://stackoverflow.com/questions/4388626/
        #  python-safe-eval-string-to-bool-int-float-none-string
        if (phy_folder / "params.py").is_file():
            with (phy_folder / "params.py").open("r") as f:
                contents = f.read()
            metadata = dict()
            contents = contents.replace("\n", " ")
            pattern = re.compile(r"(\S*)[\s]?=[\s]?(\S*)")
            elements = pattern.findall(contents)
            for key, value in elements:
                metadata[key.lower()] = ast.literal_eval(value)

        self._sampling_frequency = metadata["sample_rate"]

        clust_ids = np.unique(self._spike_clusters)
        self.unit_labels = list(clust_ids)

        self._t_start = 0.0
        self._t_stop = max(self._spike_times).item() / self._sampling_frequency

        signal_streams = []
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        signal_channels = []
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        spike_channels = []
        for i, clust_id in enumerate(clust_ids):
            unit_name = f"unit {clust_id}"
            unit_id = f"{clust_id}"
            wf_units = ""
            wf_gain = 0
            wf_offset = 0.0
            wf_left_sweep = 0
            wf_sampling_rate = 0
            spike_channels.append((unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate))
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

        csv_tsv_files = [x for x in phy_folder.iterdir() if x.suffix == ".csv" or x.suffix == ".tsv"]

        # annotation_lists is list of list of dict (python==3.8)
        # or list of list of ordered dict (python==3.6)
        # SEE: https://docs.python.org/3/library/csv.html#csv.DictReader
        annotation_lists = [self._parse_tsv_or_csv_to_list_of_dict(file) for file in csv_tsv_files]

        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["name"] = "Block #0"
        seg_ann = bl_ann["segments"][0]
        seg_ann["name"] = "Seg #0 Block #0"
        for index, clust_id in enumerate(clust_ids):
            spiketrain_an = seg_ann["spikes"][index]

            # Add cluster_id annotation
            spiketrain_an["cluster_id"] = clust_id

            # Loop over list of list of dict and annotate each st
            for annotation_list in annotation_lists:
                clust_key, *property_names = tuple(annotation_list[0].keys())
                for property_name in property_names:
                    if property_name == "KSLabel":
                        annotation_name = "quality"
                    else:
                        annotation_name = property_name.lower()
                    for annotation_dict in annotation_list:
                        if int(annotation_dict[clust_key]) == clust_id:
                            spiketrain_an[annotation_name] = annotation_dict[property_name]
                            break

            cluster_mask = (self._spike_clusters == clust_id).flatten()

            current_templates = self._spike_templates[cluster_mask].flatten()
            unique_templates = np.unique(current_templates)
            spiketrain_an["templates"] = unique_templates
            spiketrain_an["__array_annotations__"]["templates"] = current_templates

            if self._amplitudes is not None:
                spiketrain_an["__array_annotations__"]["amplitudes"] = self._amplitudes[cluster_mask]

            if self._pc_features is not None:
                current_pc_features = self._pc_features[cluster_mask]
                _, num_pcs, num_pc_channels = current_pc_features.shape
                for pc_idx in range(num_pcs):
                    for channel_idx in range(num_pc_channels):
                        key = f"channel{channel_idx}_pc{pc_idx}"
                        spiketrain_an["__array_annotations__"][key] = current_pc_features[:, pc_idx, channel_idx]

            if self._pc_feature_ind is not None:
                spiketrain_an["pc_feature_ind"] = self._pc_feature_ind[unique_templates]

    def _segment_t_start(self, block_index, seg_index):
        if block_index != 0:
            raise ValueError("`block_index` must be 0")
        return self._t_start

    def _segment_t_stop(self, block_index, seg_index):
        if block_index != 0:
            raise ValueError("`block_index` must be 0")
        return self._t_stop

    def _get_signal_size(self, block_index, seg_index, channel_indexes=None):
        return None

    def _get_signal_t_start(self, block_index, seg_index, channel_indexes):
        return None

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, channel_indexes):
        return None

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        if block_index != 0:
            raise ValueError("`block_index` must be 0")
        spikes = self._spike_clusters
        unit_label = self.unit_labels[spike_channel_index]
        mask = spikes == unit_label
        nb_spikes = np.sum(mask)
        return nb_spikes

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        if block_index != 0:
            raise ValueError("`block_index` must be 0")
        if seg_index != 0:
            raise ValueError("`seg_index` must be 0")

        unit_label = self.unit_labels[spike_channel_index]
        mask = self._spike_clusters == unit_label
        spike_timestamps = self._spike_times[mask]

        if t_start is not None:
            start_frame = int(t_start * self._sampling_frequency)
            spike_timestamps = spike_timestamps[spike_timestamps >= start_frame]
        if t_stop is not None:
            end_frame = int(t_stop * self._sampling_frequency)
            spike_timestamps = spike_timestamps[spike_timestamps < end_frame]

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self._sampling_frequency
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None

    def _event_count(self, block_index, seg_index, event_channel_index):
        return None

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        return None

    def _rescale_event_timestamp(self, event_timestamps, dtype):
        return None

    def _rescale_epoch_duration(self, raw_duration, dtype):
        return None

    @staticmethod
    def _parse_tsv_or_csv_to_list_of_dict(filename):
        list_of_dict = list()
        letter_pattern = re.compile("[a-zA-Z]")
        float_pattern = re.compile(r"-?\d*\.")
        with open(filename) as csvfile:
            if filename.suffix == ".csv":
                reader = csv.DictReader(csvfile, delimiter=",")
            elif filename.suffix == ".tsv":
                reader = csv.DictReader(csvfile, delimiter="\t")
            else:
                raise ValueError("Function parses only .csv or .tsv files")
            line = 0

            for row in reader:
                if line == 0:
                    cluster_id_key, *annotation_keys = tuple(row.keys())
                # Convert cluster ID to int
                row[cluster_id_key] = int(row[cluster_id_key])
                # Convert strings without letters
                for key in annotation_keys:
                    value = row[key]
                    if not len(value):
                        row[key] = None
                    elif letter_pattern.match(value) is None:
                        if float_pattern.match(value) is None:
                            row[key] = int(value)
                        else:
                            row[key] = float(value)

                list_of_dict.append(row)
                line += 1

        return list_of_dict



"""
Tests of neo.io.spike2io
"""

import unittest

import quantities as pq

from neo.io import Spike2IO
from neo.test.iotest.common_io_test import BaseTestIO


class TestSpike2IO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = Spike2IO
    entities_to_download = ["spike2"]
    entities_to_test = [
        "spike2/File_spike2_1.smr",
        "spike2/File_spike2_2.smr",
        "spike2/File_spike2_3.smr",
        "spike2/130322-1LY.smr",  # this is for bug 182
        "spike2/multi_sampling.smr",  # this is for bug 466
        "spike2/Two-mice-bigfile-test000.smr",  # SONv9 file
    ]

    def test_multi_sampling_no_grouping(self):
        """
        Some file can have several sampling_rate.
        This one contain 3 differents signals sampling rate
        """
        filename = self.get_local_path("spike2/multi_sampling.smr")
        reader = Spike2IO(filename=filename, try_signal_grouping=False)
        bl = reader.read_block(signal_group_mode="group-by-same-units")
        assert len(bl.segments) == 10
        seg = bl.segments[0]

        # 7 group_id one per channel
        assert len(seg.analogsignals) == 7

        # 1 channel for 1kHz
        assert seg.analogsignals[0].shape == (14296, 1)
        assert seg.analogsignals[0].sampling_rate == 1000 * pq.Hz

        # 4  channel for 2kHz
        for c in range(1, 5):
            assert seg.analogsignals[c].shape == (28632, 1)
            assert seg.analogsignals[c].sampling_rate == 2000 * pq.Hz

        # 2 channel for 10kHz
        for c in range(5, 7):
            assert seg.analogsignals[c].shape == (114618, 1)
            assert seg.analogsignals[c].sampling_rate == 10000 * pq.Hz

    def test_multi_sampling_no_grouping(self):
        """
        Some files can contain multiple sampling rates.
        This file contains three signals with different sampling rates.
        """
        filename = self.get_local_path("spike2/multi_sampling.smr")
        reader = Spike2IO(filename=filename, try_signal_grouping=True)
        bl = reader.read_block(signal_group_mode="group-by-same-units")
        assert len(bl.segments) == 10
        seg = bl.segments[0]

        # 3 groups
        assert len(seg.analogsignals) == 3

        # 1 channel for 1kHz
        assert seg.analogsignals[0].shape == (14296, 1)
        assert seg.analogsignals[0].sampling_rate == 1000 * pq.Hz

        # 4  channel for 2kHz
        assert seg.analogsignals[1].shape == (28632, 4)
        assert seg.analogsignals[1].sampling_rate == 2000 * pq.Hz

        # 2 channel for 10kHz
        assert seg.analogsignals[2].shape == (114618, 2)
        assert seg.analogsignals[2].sampling_rate == 10000 * pq.Hz


if __name__ == "__main__":
    unittest.main()



"""
Example for usecases.rst
"""

import numpy as np
from neo import Segment, AnalogSignal, SpikeTrain, Group, ChannelView
from quantities import Hz

# generate some fake data
seg = Segment()
seg.analogsignals.append(
    AnalogSignal(
        [
            [0.1, 0.1, 0.1, 0.1],
            [-2.0, -2.0, -2.0, -2.0],
            [0.1, 0.1, 0.1, 0.1],
            [-0.1, -0.1, -0.1, -0.1],
            [-0.1, -0.1, -0.1, -0.1],
            [-3.0, -3.0, -3.0, -3.0],
            [0.1, 0.1, 0.1, 0.1],
            [0.1, 0.1, 0.1, 0.1],
        ],
        sampling_rate=1000 * Hz,
        units="V",
    )
)

# extract spike trains from all channels
st_list = []
for signal in seg.analogsignals:
    # use a simple threshhold detector
    spike_mask = np.where(np.min(signal.magnitude, axis=1) < -1.0)[0]

    # create a spike train
    spike_times = signal.times[spike_mask]
    st = SpikeTrain(spike_times, t_start=signal.t_start, t_stop=signal.t_stop)

    # remember the spike waveforms
    wf_list = []
    for spike_idx in np.nonzero(spike_mask)[0]:
        wf_list.append(signal[spike_idx - 1 : spike_idx + 2, :])
    st.waveforms = np.array(wf_list)

    st_list.append(st)

unit = Group()
unit.spiketrains = st_list
unit.analogsignals.extend(seg.analogsignals)



"""
Tools for use with neo tests.
"""

import hashlib
import os

import numpy as np
import quantities as pq

import neo
from neo.core import objectlist
from neo.core.baseneo import _reference_name, _container_name
from neo.core.basesignal import BaseSignal
from neo.core.container import Container
from neo.core.objectlist import ObjectList
from neo.core.spiketrainlist import SpikeTrainList
from neo.io.basefromrawio import proxyobjectlist, EventProxy, EpochProxy


def assert_arrays_equal(a, b, dtype=False):
    """
    Check if two arrays have the same shape and contents.

    If dtype is True (default=False), then also theck that they have the same
    dtype.
    """
    assert isinstance(a, np.ndarray), f"a is a {type(a)}"
    assert isinstance(b, np.ndarray), f"b is a {type(b)}"
    assert a.shape == b.shape, f"{a} != {b}"
    # assert a.dtype == b.dtype, "%s and %s not same dtype %s %s" % (a, b,
    #                                                               a.dtype,
    #                                                               b.dtype)
    try:
        assert (a.flatten() == b.flatten()).all(), f"{a} != {b}"
    except (AttributeError, ValueError):
        try:
            ar = np.array(a)
            br = np.array(b)
            assert (ar.flatten() == br.flatten()).all(), f"{ar} != {br}"
        except (AttributeError, ValueError):
            assert np.all(a.flatten() == b.flatten()), f"{a} != {b}"

    if dtype:
        assert a.dtype == b.dtype, f"{a} and {b} not same dtype {a.dtype} and {b.dtype}"


def assert_arrays_almost_equal(a, b, threshold, dtype=False):
    """
    Check if two arrays have the same shape and contents that differ
    by abs(a - b) <= threshold for all elements.

    If threshold is None, do an absolute comparison rather than a relative
    comparison.
    """
    if threshold is None:
        return assert_arrays_equal(a, b, dtype=dtype)

    assert isinstance(a, np.ndarray), f"a is a {type(a)}"
    assert isinstance(b, np.ndarray), f"b is a {type(b)}"
    assert a.shape == b.shape, f"{a} != {b}"
    # assert a.dtype == b.dtype, "%s and %b not same dtype %s %s" % (a, b,
    #                                                               a.dtype,
    #                                                               b.dtype)
    if a.dtype.kind in ["f", "c", "i"]:
        assert (abs(a - b) < threshold).all(), (
            f"abs({a} - {b})    max(|a - b|) = {(abs(a - b)).max()}    threshold:{threshold}" ""
        )

    if dtype:
        assert a.dtype == b.dtype, f"{a} and {b} not same dtype {a.dtype} and {b.dtype}"


def file_digest(filename):
    """
    Get the sha1 hash of the file with the given filename.
    """
    with open(filename, "rb") as fobj:
        return hashlib.sha1(fobj.read()).hexdigest()


def assert_file_contents_equal(a, b):
    """
    Assert that two files have the same size and hash.
    """

    def generate_error_message(a, b):
        """
        This creates the error message for the assertion error
        """
        size_a = os.stat(a).st_size
        size_b = os.stat(b).st_size
        if size_a == size_b:
            return "Files have the same size but different contents"
        else:
            return f"Files have different sizes: a:{size_a} b: {size_b}"

    assert file_digest(a) == file_digest(b), generate_error_message(a, b)


def assert_neo_object_is_compliant(ob, check_type=True):
    """
    Test neo compliance of one object and sub objects
    (one_to_many_relation only):
      * check types and/or presence of necessary and recommended attribute.
      * If attribute is Quantities or numpy.ndarray it also check ndim.
      * If attribute is numpy.ndarray also check dtype.kind.

    check_type=True by default can be set to false for testing ProxyObject
    """
    if check_type:
        assert type(ob) in objectlist, f"{type(ob)} is not a neo object"
    classname = ob.__class__.__name__

    # test presence of necessary attributes
    for ioattr in ob._necessary_attrs:
        attrname, attrtype = ioattr[0], ioattr[1]
        # ~ if attrname != '':
        if not hasattr(ob, "_quantity_attr"):
            assert hasattr(ob, attrname), f"{classname} neo obect does not have {attrname}"

    # test attributes types
    for ioattr in ob._all_attrs:
        attrname, attrtype = ioattr[0], ioattr[1]

        if (
            hasattr(ob, "_quantity_attr")
            and ob._quantity_attr == attrname
            and (attrtype == pq.Quantity or attrtype == np.ndarray)
        ):
            # object inherits from Quantity (AnalogSignal, SpikeTrain, ...)
            ndim = ioattr[2]
            assert ob.ndim == ndim, f"{classname} dimension is {ob.ndim} should be {ndim}"
            if attrtype == np.ndarray:
                dtp = ioattr[3]
                assert ob.dtype.kind == dtp.kind, f"{classname} dtype.kind is {ob.dtype.kind} should be {dtp.kind}" ""

        elif hasattr(ob, attrname):
            if getattr(ob, attrname) is not None:
                obattr = getattr(ob, attrname)
                assert issubclass(type(obattr), attrtype), (
                    f"{attrname} in {classname} is {type(obattr)} should be {attrtype}" ""
                )
                if attrtype == pq.Quantity or attrtype == np.ndarray:
                    ndim = ioattr[2]
                    assert obattr.ndim == ndim, f"{classname}.{attrname} dimension is {obattr.ndim} should be {ndim}" ""
                if attrtype == np.ndarray:
                    dtp = ioattr[3]
                    assert obattr.dtype.kind == dtp.kind, (
                        f"{classname}.{attrname} dtype.kind is {obattr.dtype.kind} should be {dtp.kind}" ""
                    )

    # test bijectivity : parents and children
    if classname != "Group":  # objects in a Group do not keep a reference to the group.
        for container in getattr(ob, "_single_child_containers", []):
            for i, child in enumerate(getattr(ob, container, [])):
                assert hasattr(child, _reference_name(classname)), (
                    f"{container} should have {_reference_name(classname)} attribute (2 way relationship)" ""
                )
                if hasattr(child, _reference_name(classname)):
                    parent = getattr(child, _reference_name(classname))
                    assert parent == ob, (
                        f"{container}.{_reference_name(classname)} {i} is not symmetric with {classname}.{container}" ""
                    )

    # recursive on one to many rel
    for i, child in enumerate(getattr(ob, "children", [])):
        try:
            assert_neo_object_is_compliant(child)
        # intercept exceptions and add more information
        except BaseException as exc:
            exc.args += (f"from {child.__class__.__name__} {i} of {classname}",)
            raise


def types_match(ob1, ob2):
    if type(ob1) == type(ob2):
        return True
    elif isinstance(ob1, ObjectList):
        return isinstance(ob2, (list, ObjectList))
    elif isinstance(ob2, ObjectList):
        return isinstance(ob1, (list, ObjectList))
    else:
        return False


def assert_same_sub_schema(ob1, ob2, equal_almost=True, threshold=1e-10, exclude=None):
    """
    Test if ob1 and ob2 has the same sub schema.
    Explore all parent/child relationships.
    Many_to_many_relationship is not tested
    because of infinite recursive loops.

    Arguments:
        equal_almost: if False do a strict arrays_equal if
                      True do arrays_almost_equal
        exclude: a list of attributes and annotations to ignore in
                 the comparison

    """
    if isinstance(ob1, SpikeTrainList) and isinstance(ob2, list):
        # for debugging occasional test failure
        raise Exception(
            f"items={str(ob1._items)}\nspike_time_array={str(ob1._spike_time_array)}\nlist length: {len(ob2)}"
        )
    errmsg = f"type({type(ob1)}) != type({type(ob2)})"
    assert types_match(ob1, ob2), errmsg
    classname = ob1.__class__.__name__

    if exclude is None:
        exclude = []

    if isinstance(ob1, (list, ObjectList)):
        assert len(ob1) == len(ob2), f"lens {len(ob1)} and {len(ob2)} not equal for {ob1} and {ob2}" ""
        for i, (sub1, sub2) in enumerate(zip(ob1, ob2)):
            try:
                assert_same_sub_schema(sub1, sub2, equal_almost=equal_almost, threshold=threshold, exclude=exclude)
            # intercept exceptions and add more information
            except BaseException as exc:
                exc.args += f"{classname}[{i}]"
                raise
        return

    # test parent/child relationship
    for container in getattr(ob1, "_single_child_containers", []):
        if container in exclude:
            continue
        if not hasattr(ob1, container):
            assert not hasattr(ob2, container), f"{classname} 2 does have {container} but not {classname} 1" ""
            continue
        else:
            assert hasattr(ob2, container), f"{classname} 1 has {container} but not {classname} 2"

        sub1 = getattr(ob1, container)
        sub2 = getattr(ob2, container)

        assert len(sub1) == len(sub2), (
            f"these two {classname} do not have the same {container} number: {len(sub1)} and {len(sub2)}" ""
        )
        for i in range(len(getattr(ob1, container))):
            # previously lacking parameter
            try:
                assert_same_sub_schema(
                    sub1[i], sub2[i], equal_almost=equal_almost, threshold=threshold, exclude=exclude
                )
            # intercept exceptions and add more information
            except BaseException as exc:
                exc.args += f"from {container}[{i}] of {classname}"
                raise

    assert_same_attributes(ob1, ob2, equal_almost=equal_almost, threshold=threshold, exclude=exclude)


def assert_same_attributes(ob1, ob2, equal_almost=True, threshold=1e-10, exclude=None):
    """
    Test if ob1 and ob2 has the same attributes.

    Arguments:
        equal_almost: if False do a strict arrays_equal if
                      True do arrays_almost_equal
        exclude: a list of attributes and annotations to ignore in
                 the comparison

    """
    classname = ob1.__class__.__name__

    if exclude is None:
        exclude = []

    if not equal_almost:
        threshold = None
        dtype = True
    else:
        dtype = False

    for ioattr in ob1._all_attrs:
        if ioattr[0] in exclude:
            continue
        attrname, attrtype = ioattr[0], ioattr[1]
        # ~ if attrname =='':
        if hasattr(ob1, "_quantity_attr") and ob1._quantity_attr == attrname:
            # object is inherited from Quantity (AnalogSignal, SpikeTrain, ...)
            try:
                assert_arrays_almost_equal(ob1.magnitude, ob2.magnitude, threshold=threshold, dtype=dtype)
            # intercept exceptions and add more information
            except BaseException as exc:
                exc.args += (f"from {classname} {attrname}",)
                raise
            assert ob1.dimensionality.string == ob2.dimensionality.string, (
                f"Units of {classname} {attrname} are not the same: {ob1.dimensionality.string} and {ob2.dimensionality.string}"
                ""
            )
            continue

        if not hasattr(ob1, attrname):
            assert not hasattr(ob2, attrname), f"{classname} 2 does have {attrname} but not {classname} 1" ""
            continue
        else:
            assert hasattr(ob2, attrname), f"%{classname} 1 has {attrname} but not {classname} 2" ""

        if getattr(ob1, attrname) is None:
            assert getattr(ob2, attrname) is None, (
                f"In {classname}.{attrname} {getattr(ob1, attrname)} and {getattr(ob2, attrname)} differed" ""
            )
            continue
        if getattr(ob2, attrname) is None:
            assert getattr(ob1, attrname) is None, (
                f"In {classname}.{attrname} {getattr(ob1, attrname)} and { getattr(ob2, attrname)} differed" ""
            )
            continue

        if attrtype == pq.Quantity:
            # Compare magnitudes
            mag1 = getattr(ob1, attrname).magnitude
            mag2 = getattr(ob2, attrname).magnitude
            # print "2. ob1(%s) %s:%s\n   ob2(%s) %s:%s" % \
            # (ob1,attrname,mag1,ob2,attrname,mag2)
            try:
                assert_arrays_almost_equal(mag1, mag2, threshold=threshold, dtype=dtype)
            # intercept exceptions and add more information
            except BaseException as exc:
                exc.args += (f"from {attrname} of {classname}",)
                raise
            # Compare dimensionalities
            dim1 = getattr(ob1, attrname).dimensionality.simplified
            dim2 = getattr(ob2, attrname).dimensionality.simplified
            dimstr1 = getattr(ob1, attrname).dimensionality.string
            dimstr2 = getattr(ob2, attrname).dimensionality.string
            assert dim1 == dim2, f"Attribute {attrname} of {classname} are not the same: {dimstr1} != {dimstr2}" ""
        elif attrtype == np.ndarray:
            try:
                assert_arrays_almost_equal(
                    getattr(ob1, attrname), getattr(ob2, attrname), threshold=threshold, dtype=dtype
                )
            # intercept exceptions and add more information
            except BaseException as exc:
                exc.args += (f"from {attrname} of {classname}",)
                raise

        elif isinstance(attrtype, tuple):
            attr1 = getattr(ob1, attrname)
            attr2 = getattr(ob2, attrname)
            assert attr1.__class__.__name__ in attrtype
            assert attr2.__class__.__name__ in attrtype
            if isinstance(attr1, BaseSignal):
                assert isinstance(attr2, BaseSignal)
                # Compare magnitudes
                assert_arrays_almost_equal(attr1.magnitude, attr2.magnitude, threshold=threshold, dtype=dtype)
                # Compare dimensionalities
                dim1 = attr1.dimensionality
                dim2 = attr2.dimensionality
                errmsg = f"Attribute {attrname} of {classname} are not the same: {dim1.string} != {dim2.string}"
                assert dim1.simplified == dim2.simplified, errmsg
                assert attr1.name == attr2.name
                # todo: check annotations
        else:
            # ~ print 'yep', getattr(ob1, attrname),  getattr(ob2, attrname)
            assert getattr(ob1, attrname) == getattr(ob2, attrname), (
                f"Attribute {classname}.{attrname} are not the same { type(getattr(ob1, attrname))} {getattr(ob1, attrname)} {type(getattr(ob2, attrname))} {getattr(ob2, attrname)}"
                ""
            )


def assert_same_annotations(ob1, ob2, equal_almost=True, threshold=1e-10, exclude=None):
    """
    Test if ob1 and ob2 has the same annotations.

    Arguments:
        equal_almost: if False do a strict arrays_equal if
                      True do arrays_almost_equal
        exclude: a list of attributes and annotations to ignore in
                 the comparison

    """
    if exclude is None:
        exclude = []

    if not equal_almost:
        threshold = None
        dtype = False
    else:
        dtype = True

    for key in ob2.annotations:
        if key in exclude:
            continue
        assert key in ob1.annotations

    for key, value1 in ob1.annotations.items():
        if key in exclude:
            continue
        assert key in ob2.annotations
        value2 = ob2.annotations[key]
        if isinstance(value1, np.ndarray):
            assert isinstance(value2, np.ndarray)
            assert_arrays_almost_equal(value1, value2, threshold=threshold, dtype=False)
        else:
            assert value1 == value2


def assert_same_array_annotations(ob1, ob2, equal_almost=True, threshold=1e-10, exclude=None):
    """
    Test if ob1 and ob2 has the same annotations.

    Arguments:
        equal_almost: if False do a strict arrays_equal if
                      True do arrays_almost_equal
        exclude: a list of attributes and annotations to ignore in
                 the comparison

    """
    if exclude is None:
        exclude = []

    if not equal_almost:
        threshold = None
        dtype = False
    else:
        dtype = True

    for key in ob2.array_annotations:
        if key in exclude:
            continue
        assert key in ob1.array_annotations

    for key, value in ob1.array_annotations.items():
        if key in exclude:
            continue
        assert key in ob2.array_annotations
        try:
            assert_arrays_equal(value, ob2.array_annotations[key])
        except ValueError:
            assert_arrays_almost_equal(ob1, ob2, threshold=threshold, dtype=False)


def assert_sub_schema_is_lazy_loaded(ob):
    """
    This is util for testing lazy load. All data object must be in proxyobjectlist.
    """
    classname = ob.__class__.__name__

    if isinstance(ob, Container):
        for container in getattr(ob, "_single_child_containers", []):
            if not hasattr(ob, container):
                continue
            sub = getattr(ob, container)
            for i, child in enumerate(sub):
                try:
                    assert_sub_schema_is_lazy_loaded(child)
                # intercept exceptions and add more information
                except BaseException as exc:
                    exc.args += (f"from {container} {i} of {classname}",)
                    raise
    else:
        assert ob.__class__ in proxyobjectlist, f"Data object must lazy {classname}"
        loaded_ob = ob.load()
        assert_neo_object_is_compliant(loaded_ob)
        assert_same_annotations(ob, loaded_ob)
        exclude = []
        if isinstance(ob, EventProxy):
            exclude = ["labels"]
        elif isinstance(ob, EpochProxy):
            exclude = ["labels", "durations"]
        else:
            exclude = []
        assert_same_array_annotations(ob, loaded_ob, exclude=exclude)


def assert_objects_equivalent(obj1, obj2):
    """
    Compares two NEO objects by looping over the attributes and annotations
    and asserting their hashes. No relationships involved.
    """

    def assert_attr(obj1, obj2, attr_name):
        """
        Assert a single attribute and annotation are the same
        """
        assert hasattr(obj1, attr_name)
        attr1 = hashlib.md5(getattr(obj1, attr_name)).hexdigest()
        assert hasattr(obj2, attr_name)
        attr2 = hashlib.md5(getattr(obj2, attr_name)).hexdigest()
        assert attr1 == attr2, f"Attribute {attr_name} for class {obj1.__class__.__name__} is not equal." ""

    obj_type = obj1.__class__.__name__
    assert obj_type == obj2.__class__.__name__
    for ioattr in obj1._necessary_attrs:
        assert_attr(obj1, obj2, ioattr[0])
    for ioattr in obj1._recommended_attrs:
        if hasattr(obj1, ioattr[0]) or hasattr(obj2, ioattr[0]):
            assert_attr(obj1, obj2, ioattr[0])
    if hasattr(obj1, "annotations"):
        assert hasattr(obj2, "annotations")
        for key, value in obj1.annotations:
            assert hasattr(obj2.annotations, key)
            assert obj2.annotations[key] == value


def assert_children_empty(obj, parent):
    """
    Check that the children of a neo object are empty.  Used
    to check the cascade is implemented properly
    """
    classname = obj.__class__.__name__
    errmsg = f"""{parent.__name__} reader with cascade=False should return
        empty children"""
    if hasattr(obj, "children"):
        assert not obj.children, errmsg



# -*- coding: utf-8 -*-
"""
This module implements :class:`SpikeTrainList`, a pseudo-list
which supports a multiplexed representation of spike trains
(all times in a single array, with a second array indicating which
neuron/channel the spike is from).
"""

import warnings

import numpy as np
import quantities as pq

from .spiketrain import SpikeTrain, normalize_times_array
from .objectlist import ObjectList


def is_spiketrain_or_proxy(obj):
    return isinstance(obj, SpikeTrain) or getattr(obj, "proxy_for", None) == SpikeTrain


def unique(quantities):
    """
    Function to mimic np.unique for use with quantities

    Parameters
    ----------
    quantities: array-like[pq.Quantity dtype]
        An array-like object containing quantities

    Returns
    -------
    unique_quantities: array[pq.Quantity dtype]
        The unique entries within the array
    Notes
    -----
    np.unique doesn't work with a list of quantities of different scale,
    this function can be used instead."""
    # todo: add a tolerance to handle floating point discrepancies
    #       due to scaling.
    if len(quantities) > 0:
        common_units = quantities[0].units
        scaled_quantities = pq.Quantity([q.rescale(common_units) for q in quantities], common_units)
        return np.unique(scaled_quantities)
    else:
        return quantities


class SpikeTrainList(ObjectList):
    """
    This class contains multiple spike trains, and can represent them
    either as a list of SpikeTrain objects or as a pair of arrays
    (all spike times in a single array, with a second array indicating which
    neuron/channel the spike is from).

    Parameters
    ----------
    items: list[neo.SpikeTrain] | array[neo.SpikeTrain] | None, default: None
        The SpikeTrains to add to the list
    parent: neo.Segment | None, default: None
        The Segment to include the SpikeTrainList within or None for no parent

    Notes
    -----
    A SpikeTrainList object should behave like a list of SpikeTrains
    for iteration and item access. It is not intended to be used directly
    by users, but is available as the attribute `spiketrains` of Segments.

    Examples
    --------
    # Create from list of SpikeTrain objects

    >>> stl = SpikeTrainList(items=(
    ...     SpikeTrain([0.5, 0.6, 23.6, 99.2], units="ms", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
    ...     SpikeTrain([0.0007, 0.0112], units="s", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
    ...     SpikeTrain([1100, 88500], units="us", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
    ...     SpikeTrain([], units="ms", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
    ...                             )
    ...                     )
    >>> stl.multiplexed
    (array([0, 0, 0, 0, 1, 1, 2, 2]),
    array([ 0.5,  0.6, 23.6, 99.2,  0.7, 11.2,  1.1, 88.5]) * ms)

    # Create from a pair of arrays

    >>> stl = SpikeTrainList.from_spike_time_array(
    ...     np.array([0.5, 0.6, 0.7, 1.1, 11.2, 23.6, 88.5, 99.2]),
    ...     np.array([0, 0, 1, 2, 1, 0, 2, 0]),
    ...     all_channel_ids=[0, 1, 2, 3],
    ...     units='ms',
    ...     t_start=0 * pq.ms,
    ...     t_stop=100.0 * pq.ms)
    >>> list(stl)
    [<SpikeTrain(array([ 0.5,  0.6, 23.6, 99.2]) * ms, [0.0 ms, 100.0 ms])>,
    <SpikeTrain(array([ 0.7, 11.2]) * ms, [0.0 ms, 100.0 ms])>,
    <SpikeTrain(array([ 1.1, 88.5]) * ms, [0.0 ms, 100.0 ms])>,
    <SpikeTrain(array([], dtype=float64) * ms, [0.0 ms, 100.0 ms])>]

    """

    allowed_contents = (SpikeTrain,)

    def __init__(self, items=None, parent=None):
        """Initialize self"""
        if items is None:
            self._items = items
        else:
            for item in items:
                if not is_spiketrain_or_proxy(item):
                    raise ValueError("`items` can only contain SpikeTrain objects or proxy objects")
            self._items = list(items)
        self._spike_time_array = None
        self._channel_id_array = None
        self._all_channel_ids = None
        self._spiketrain_metadata = {}
        if parent is not None and parent.__class__.__name__ != "Segment":
            raise AttributeError("The parent class must be a Segment")
        self.segment = parent

    @property
    def parent(self):
        return self.segment

    def __iter__(self):
        """Implement iter(self)"""
        if self._items is None:
            self._spiketrains_from_array()
        for item in self._items:
            yield item

    def __getitem__(self, i):
        """x.__getitem__(y) <==> x[y]"""
        if self._items is None:
            self._spiketrains_from_array()
        items = self._items[i]
        if is_spiketrain_or_proxy(items):
            return items
        else:
            return SpikeTrainList(items=items)

    def __setitem__(self, i, value):
        if self._items is None:
            self._spiketrains_from_array()
        self._items[i] = value

    def __str__(self):
        """Return str(self)"""
        if self._items is None:
            if self._spike_time_array is None:
                return str([])
            else:
                return (
                    f"SpikeTrainList containing {self._spike_time_array.size} "
                    f"spikes from {len(self._all_channel_ids)} neurons"
                )
        else:
            return str(self._items)

    def __repr__(self):
        return "<SpikeTrainList>"

    def __len__(self):
        """Return len(self)"""
        if self._items is None:
            if self._all_channel_ids is not None:
                return len(self._all_channel_ids)
            elif self._channel_id_array is not None:
                return np.unique(self._channel_id_array).size
            else:
                return 0
        else:
            return len(self._items)

    def _add_spiketrainlists(self, other, in_place=False):
        if self._spike_time_array is None or other._spike_time_array is None:
            # if either self or other is not storing multiplexed spike trains
            # we combine them using the list of SpikeTrains representation
            if self._items is None:
                self._spiketrains_from_array()
            if other._items is None:
                other._spiketrains_from_array()
            if in_place:
                self._items.extend(other._items)
                return self
            else:
                return self.__class__(items=self._items[:] + other._items)
        else:
            # both self and other are storing multiplexed spike trains
            # so we update the array representation
            if self._spiketrain_metadata["t_start"] != other._spiketrain_metadata["t_start"]:
                raise ValueError("Incompatible t_start")
                # todo: adjust times and t_start of other to be compatible with self
            if self._spiketrain_metadata["t_stop"] != other._spiketrain_metadata["t_stop"]:
                raise ValueError("Incompatible t_stop")
                # todo: adjust t_stop of self and other as necessary
            combined_spike_time_array = np.hstack((self._spike_time_array, other._spike_time_array))
            combined_channel_id_array = np.hstack((self._channel_id_array, other._channel_id_array))
            combined_channel_ids = set(list(self._all_channel_ids) + other._all_channel_ids)
            if len(combined_channel_ids) != (len(self._all_channel_ids) + len(other._all_channel_ids)):
                raise ValueError("Duplicate channel ids, please rename channels before adding")
            if in_place:
                self._spike_time_array = combined_spike_time_array
                self._channel_id_array = combined_channel_id_array
                self._all_channel_ids = combined_channel_ids
                self._items = None
                return self
            else:
                return self.__class__.from_spike_time_array(
                    combined_spike_time_array,
                    combined_channel_id_array,
                    combined_channel_ids,
                    t_start=self._spiketrain_metadata["t_start"],
                    t_stop=self._spiketrain_metadata["t_stop"],
                )

    def __add__(self, other):
        """Return self + other"""
        if isinstance(other, self.__class__):
            return self._add_spiketrainlists(other)
        elif other and is_spiketrain_or_proxy(other[0]):
            return self._add_spiketrainlists(self.__class__(items=other, parent=self.segment))
        else:
            if self._items is None:
                self._spiketrains_from_array()
            return self._items + other

    def __iadd__(self, other):
        """Return self"""
        if isinstance(other, self.__class__):
            return self._add_spiketrainlists(other, in_place=True)
        elif other and is_spiketrain_or_proxy(other[0]):
            for obj in other:
                self._handle_append(obj)
            if self._items is None:
                self._spiketrains_from_array()
            self._items.extend(other)
            return self
        else:
            raise TypeError("Can only add a SpikeTrainList or a list of SpikeTrains in place")

    def __radd__(self, other):
        """Return other + self"""
        if isinstance(other, self.__class__):
            return other._add_spiketrainlists(self)
        elif other and is_spiketrain_or_proxy(other[0]):
            for obj in other:
                obj.segment = self.segment
            if self._items is None:
                self._spiketrains_from_array()
            self._items.extend(other)
            return self
        elif len(other) == 0:
            return self
        else:
            if self._items is None:
                self._spiketrains_from_array()
            return other + self._items

    def append(self, obj):
        """
        Appends to the SpikeTrainList with a new neo.core.SpikeTrain

        Parameters
        ----------
        obj: neo.core.SpikeTrain
            The new neo.core.SpikeTrain to append to the current SpikeTrainList

        Examples
        --------
        # with SpikeTrainList stl with two SpikeTrains
        >>> len(stl)
        2
        >>> stl.append(SpikeTrain([0.5, 0.6, 23.6, 99.2], units="ms", t_start=0 * pq.ms, t_stop=100.0 * pq.ms))
        >>> len(stl)
        3
        """
        if not is_spiketrain_or_proxy(obj):
            raise ValueError("Can only append SpikeTrain objects")
        if self._items is None:
            self._spiketrains_from_array()
        self._handle_append(obj)
        self._items.append(obj)

    def extend(self, iterable):
        """Extends the SpikeTrainList with additional SpikeTrain's from an iterable

        Parameters
        ----------
        iterable: iterable[neo.core.SpikeTrain]
            A list-like or array-like object containing neo.core.SpikeTrain to be added to the SpikeTrainList

        Examples
        --------
        # with SpikeTrainList stl with two SpikeTrains and stl_other with three SpikeTrains
        >>> len(stl)
        2
        >>> len(stl_other)
        3
        >>> stl.extend(stl_other)
        >>> len(stl)
        5
        """

        if self._items is None:
            self._spiketrains_from_array()
        for obj in iterable:
            self._handle_append(obj)
        self._items.extend(iterable)

    @classmethod
    def from_spike_time_array(
        cls, spike_time_array, channel_id_array, all_channel_ids, t_stop, units=None, t_start=0.0 * pq.s, **annotations
    ):
        """Create a SpikeTrainList object from an array of spike times
        and an array of channel ids.

        Parameters
        ----------
        spike_time_array: quantity array 1D | numpy array 1D, | list
            The times of all spikes.
        channel_id_array: numpy array 1D of dtype int
            The id of the channel (e.g. the neuron) to which each spike belongs. This array should have the same length
            as the `spike_time_array`
        all_channel_ids: list | tuple, | numpy array 1D containing integers
            All channel ids. This is needed to represent channels in which there are no spikes.
        t_stop: quantity scalar | numpy scalar | float
            Time at which spike recording ended. This will be converted to the same units as `spike_time_array` or `units`.
        units: quantity units | None, default: None
            Required if `spike_time_array is not a :class:`~quantities.Quantity`.
        :t_start: quantity scalar | numpy scalar | float, default: 0.0 * pq.s
            Time at which spike recording began. This will be converted to the same units as `spike_time_array` or `units`
        **annotations: dict
            The neo.core.baseneo annotations (e.g. name, or user-defined)

        """
        spike_time_array, dim = normalize_times_array(spike_time_array, units)
        obj = cls()
        obj._spike_time_array = spike_time_array
        obj._channel_id_array = channel_id_array
        obj._all_channel_ids = all_channel_ids
        obj._spiketrain_metadata = {"t_start": t_start, "t_stop": t_stop}
        for name, ann_value in annotations.items():
            if (
                not isinstance(ann_value, str)
                and hasattr(ann_value, "__len__")
                and len(ann_value) != len(all_channel_ids)
            ):
                raise ValueError(f"incorrect length for annotation '{name}'")
        obj._annotations = annotations
        return obj

    def _spiketrains_from_array(self):
        """Convert multiplexed spike time data into a list of SpikeTrain objects"""
        if self._spike_time_array is None:
            self._items = []
        else:
            self._items = []
            for i, channel_id in enumerate(self._all_channel_ids):
                mask = self._channel_id_array == channel_id
                times = self._spike_time_array[mask]
                spiketrain = SpikeTrain(times, **self._spiketrain_metadata)
                for name, value in self._annotations.items():
                    if (
                        not isinstance(value, str)
                        and hasattr(value, "__len__")
                        and len(value) == len(self._all_channel_ids)
                    ):
                        spiketrain.annotate(**{name: value[i]})
                    else:
                        spiketrain.annotate(**{name: value})
                spiketrain.annotate(channel_id=channel_id)
                spiketrain.segment = self.segment
                self._items.append(spiketrain)

    @property
    def multiplexed(self):
        """
        Return spike trains as a pair of arrays.

        Returns
        -------
        channel_id_array: np.ndarray
            The ids of the channels/neurons that produced each spike
        spike_time_array: np.ndarray of dtype Quantity
            The Quantity array containing the times of the spikes
        """
        if self._spike_time_array is None:
            # need to convert list of SpikeTrains into multiplexed spike times array
            if self._items is None:
                return np.array([]), np.array([])
            else:
                channel_ids = []
                spike_times = []
                dim = self._items[0].units.dimensionality
                for i, spiketrain in enumerate(self._items):
                    if hasattr(spiketrain, "load"):  # proxy object
                        spiketrain = spiketrain.load()
                    if spiketrain.times.dimensionality.items() == dim.items():
                        # no need to rescale
                        spike_times.append(spiketrain.times)
                    else:
                        spike_times.append(spiketrain.times.rescale(dim))
                    if "channel_id" in spiketrain.annotations and isinstance(spiketrain.annotations["channel_id"], int):
                        ch_id = spiketrain.annotations["channel_id"]
                    else:
                        ch_id = i
                    channel_ids.append(ch_id * np.ones(spiketrain.shape, dtype=np.int64))
                self._spike_time_array = np.hstack(spike_times) * self._items[0].units
                self._channel_id_array = np.hstack(channel_ids)
        return self._channel_id_array, self._spike_time_array

    @property
    def t_start(self):
        if "t_start" in self._spiketrain_metadata:
            return self._spiketrain_metadata["t_start"]
        else:
            t_start_values = unique(
                [item.t_start for item in self._items if isinstance(item, SpikeTrain)]
            )  # ignore proxy objects
            if len(t_start_values) == 0:
                raise ValueError("t_start not defined for an empty spike train list")
            elif len(t_start_values) > 1:
                warnings.warn("Found multiple values of t_start, returning the earliest")
                t_start = t_start_values.min()
            else:
                t_start = t_start_values[0]
            self._spiketrain_metadata["t_start"] = t_start
            return t_start

    @property
    def t_stop(self):
        if "t_stop" in self._spiketrain_metadata:
            return self._spiketrain_metadata["t_stop"]
        else:
            t_stop_values = unique(
                [item.t_stop for item in self._items if isinstance(item, SpikeTrain)]
            )  # ignore proxy objects
            if len(t_stop_values) == 0:
                raise ValueError("t_stop not defined for an empty spike train list")
            elif len(t_stop_values) > 1:
                warnings.warn("Found multiple values of t_stop, returning the latest")
                t_stop = t_stop_values.max()
            else:
                t_stop = t_stop_values[0]
            self._spiketrain_metadata["t_stop"] = t_stop
            return t_stop

    @property
    def all_channel_ids(self):
        if self._all_channel_ids is None:
            self._all_channel_ids = []
            for i, spiketrain in enumerate(self._items):
                if "channel_id" in spiketrain.annotations and isinstance(spiketrain.annotations["channel_id"], int):
                    ch_id = spiketrain.annotations["channel_id"]
                else:
                    ch_id = i
                self._all_channel_ids.append(ch_id)
        return self._all_channel_ids



"""
Class for reading data from a SpikeGLX system  (NI-DAQ for neuropixel probe)
See https://billkarsh.github.io/SpikeGLX/


Here an adaptation of the spikeglx tools into the neo rawio API.

Note that each pair of ".bin"/."meta" files is represented as a stream of channels
that share the same sampling rate.
It will be one AnalogSignal multi channel at neo.io level.

Contrary to other implementations this IO reads the entire folder and subfolder and:
  * deals with severals segment based on the `_gt0`, `_gt1`, `_gt2`, etc postfixes
  * deals with all signals "imec0", "imec1" for neuropixel probes and also
     external signal like"nidq". This is the "device"
  * For imec device both "ap" and "lf" are extracted so one device have several "streams"

Note:
  * there are several versions depending the neuropixel probe generation (`1.x`/`2.x`/`3.x`)
    Here, we assume that the `meta` file has the same structure across all generations.
    This need so be checked.
    This IO is developed based on neuropixel generation 2.0, single shank recordings.


# Not implemented yet in this reader:
  * contact SpkeGLX developer to see how to deal with absolute t_start when several segment
  * contact SpkeGLX developer to understand the last channel SY0 function
  * better handling of annotations at object level by sub group of device (after rawio change)
  * better handling of channel location


See:
https://billkarsh.github.io/SpikeGLX/
https://billkarsh.github.io/SpikeGLX/#offline-analysis-tools
https://billkarsh.github.io/SpikeGLX/#metadata-guides
https://github.com/SpikeInterface/spikeextractors/blob/master/spikeextractors/extractors/spikeglxrecordingextractor/spikeglxrecordingextractor.py

This reader handle:

imDatPrb_type=1 (NP 1.0)
imDatPrb_type=21 (NP 2.0, single multiplexed shank)
imDatPrb_type=24 (NP 2.0, 4-shank)
imDatPrb_type=1030 (NP 1.0-NHP 45mm SOI90 - NHP long 90um wide, staggered contacts)
imDatPrb_type=1031 (NP 1.0-NHP 45mm SOI125 - NHP long 125um wide, staggered contacts)
imDatPrb_type=1032 (NP 1.0-NHP 45mm SOI115 / 125 linear - NHP long 125um wide, linear contacts)
imDatPrb_type=1022 (NP 1.0-NHP 25mm - NHP medium)
imDatPrb_type=1015 (NP 1.0-NHP 10mm - NHP short)

Author : Samuel Garcia
Some functions are copied from Graham Findlay
"""

from pathlib import Path
import os
import re

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class SpikeGLXRawIO(BaseRawIO):
    """
    Class for reading data from a SpikeGLX system

    Parameters
    ----------
    dirname: str, default: ''
        The spikeglx folder containing meta/bin files
    load_sync_channel: bool, default: False
        The last channel (SY0) of each stream is a fake channel used for synchronisation
    load_channel_location: bool, default: False
        If True probeinterface is used to load the channel locations from the directory

    Notes
    -----
    * Contrary to other implementations this IO reads the entire folder and subfolders and:
      deals with several segments based on the `_gt0`, `_gt1`, `_gt2`, etc postfixes
      deals with all signals "imec0", "imec1" for neuropixel probes and also
      external signal like"nidq". This is the "device"
    * For imec device both "ap" and "lf" are extracted so one device have several "streams"
    * There are several versions depending the neuropixel probe generation (`1.x`/`2.x`/`3.x`)
    * Here, we assume that the `meta` file has the same structure across all generations.
    * This IO is developed based on neuropixel generation 2.0, single shank recordings.

    Examples
    --------
    >>> import neo.rawio
    >>> reader = neo.rawio.SpikeGLXRawIO(dirname='path/to/data/')
    >>> reader.parse_header()
    >>> print(reader)
    >>> raw_chunk = reader.get_analogsignal_chunk(block_index=0,
                                                  seg_index=0,
                                                  i_start=None,
                                                  i_stop=None,
                                                  stream_index=0)
    """

    # file formats used by spikeglxio
    extensions = ["meta", "bin"]
    rawmode = "one-dir"

    def __init__(self, dirname="", load_sync_channel=False, load_channel_location=False):
        BaseRawIO.__init__(self)
        self.dirname = dirname
        self.load_sync_channel = load_sync_channel
        self.load_channel_location = load_channel_location

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        self.signals_info_list = scan_files(self.dirname)

        # sort stream_name by higher sampling rate first
        srates = {info["stream_name"]: info["sampling_rate"] for info in self.signals_info_list}
        stream_names = sorted(list(srates.keys()), key=lambda e: srates[e])[::-1]
        nb_segment = np.unique([info["seg_index"] for info in self.signals_info_list]).size

        self._memmaps = {}
        self.signals_info_dict = {}
        for info in self.signals_info_list:
            # key is (seg_index, stream_name)
            key = (info["seg_index"], info["stream_name"])
            if key in self.signals_info_dict:
                raise KeyError(f"key {key} is already in the signals_info_dict")
            self.signals_info_dict[key] = info

            # create memmap
            data = np.memmap(info["bin_file"], dtype="int16", mode="r", offset=0, order="C")
            # this should be (info['sample_length'], info['num_chan'])
            # be some file are shorten
            data = data.reshape(-1, info["num_chan"])
            self._memmaps[key] = data

        # create channel header
        signal_streams = []
        signal_channels = []
        for stream_name in stream_names:
            # take first segment
            info = self.signals_info_dict[0, stream_name]

            stream_id = stream_name
            stream_index = stream_names.index(info["stream_name"])
            signal_streams.append((stream_name, stream_id))

            # add channels to global list
            for local_chan in range(info["num_chan"]):
                chan_name = info["channel_names"][local_chan]
                chan_id = f"{stream_name}#{chan_name}"
                signal_channels.append(
                    (
                        chan_name,
                        chan_id,
                        info["sampling_rate"],
                        "int16",
                        info["units"],
                        info["channel_gains"][local_chan],
                        info["channel_offsets"][local_chan],
                        stream_id,
                    )
                )
            # check sync channel validity
            if "nidq" not in stream_name:
                if not self.load_sync_channel and info["has_sync_trace"]:
                    signal_channels = signal_channels[:-1]
                if self.load_sync_channel and not info["has_sync_trace"]:
                    raise ValueError("SYNC channel is not present in the recording. " "Set load_sync_channel to False")

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        # This is true only in case of 'nidq' stream
        for stream_name in stream_names:
            if "nidq" in stream_name:
                info = self.signals_info_dict[0, stream_name]
                if len(info["digital_channels"]) > 0:
                    # add event channels
                    for local_chan in info["digital_channels"]:
                        chan_name = local_chan
                        chan_id = f"{stream_name}#{chan_name}"
                        event_channels.append((chan_name, chan_id, "event"))
                    # add events_memmap
                    data = np.memmap(info["bin_file"], dtype="int16", mode="r", offset=0, order="C")
                    data = data.reshape(-1, info["num_chan"])
                    # The digital word is usually the last channel, after all the individual analog channels
                    extracted_word = data[:, len(info["analog_channels"])]
                    self._events_memmap = np.unpackbits(extracted_word.astype(np.uint8)[:, None], axis=1)
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # deal with nb_segment and t_start/t_stop per segment
        self._t_starts = {seg_index: 0.0 for seg_index in range(nb_segment)}
        self._t_stops = {seg_index: 0.0 for seg_index in range(nb_segment)}
        for seg_index in range(nb_segment):
            for stream_name in stream_names:
                info = self.signals_info_dict[seg_index, stream_name]
                t_stop = info["sample_length"] / info["sampling_rate"]
                self._t_stops[seg_index] = max(self._t_stops[seg_index], t_stop)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()
        self._generate_minimal_annotations()
        block_ann = self.raw_annotations["blocks"][0]

        for seg_index in range(nb_segment):
            seg_ann = self.raw_annotations["blocks"][0]["segments"][seg_index]
            seg_ann["name"] = f"Segment {seg_index}"

            for c, signal_stream in enumerate(signal_streams):
                stream_name = signal_stream["name"]
                sig_ann = self.raw_annotations["blocks"][0]["segments"][seg_index]["signals"][c]

                if self.load_channel_location:
                    # need probeinterface to be installed
                    import probeinterface

                    info = self.signals_info_dict[seg_index, stream_name]
                    if "imroTbl" in info["meta"] and info["stream_kind"] == "ap":
                        # only for ap channel
                        probe = probeinterface.read_spikeglx(info["meta_file"])
                        loc = probe.contact_positions
                        if self.load_sync_channel:
                            # one fake channel  for "sys0"
                            loc = np.concatenate((loc, [[0.0, 0.0]]), axis=0)
                        for ndim in range(loc.shape[1]):
                            sig_ann["__array_annotations__"][f"channel_location_{ndim}"] = loc[:, ndim]

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stops[seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        memmap = self._memmaps[seg_index, stream_id]
        return int(memmap.shape[0])

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        memmap = self._memmaps[seg_index, stream_id]
        stream_name = self.header["signal_streams"]["name"][stream_index]

        # take care of sync channel
        info = self.signals_info_dict[0, stream_name]
        if not self.load_sync_channel and info["has_sync_trace"]:
            memmap = memmap[:, :-1]

        # since we cut the memmap, we can simplify the channel selection
        if channel_indexes is None:
            channel_selection = slice(None)
        elif isinstance(channel_indexes, slice):
            channel_selection = channel_indexes
        elif not isinstance(channel_indexes, slice):
            if np.all(np.diff(channel_indexes) == 1):
                # consecutive channel then slice this avoid a copy (because of ndarray.take(...)
                # and so keep the underlying memmap
                channel_selection = slice(channel_indexes[0], channel_indexes[0] + len(channel_indexes))
            else:
                channel_selection = channel_indexes
        else:
            raise ValueError("get_analogsignal_chunk : channel_indexes" "must be slice or list or array of int")

        raw_signals = memmap[slice(i_start, i_stop), channel_selection]

        return raw_signals

    def _event_count(self, event_channel_idx, block_index=None, seg_index=None):
        timestamps, _, _ = self._get_event_timestamps(block_index, seg_index, event_channel_idx, None, None)
        return timestamps.size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start=None, t_stop=None):
        timestamps, durations, labels = [], None, []
        info = self.signals_info_dict[0, "nidq"]  # There are no events that are not in the nidq stream
        dig_ch = info["digital_channels"]
        if len(dig_ch) > 0:
            event_data = self._events_memmap
            channel = dig_ch[event_channel_index]
            ch_idx = 7 - int(channel[2:])  # They are in the reverse order
            this_stream = event_data[:, ch_idx]
            this_rising = np.where(np.diff(this_stream) == 1)[0] + 1
            this_falling = (
                np.where(np.diff(this_stream) == 255)[0] + 1
            )  # because the data is in unsigned 8 bit, -1 = 255!
            if len(this_rising) > 0:
                timestamps.extend(this_rising)
                labels.extend([f"{channel} ON"] * len(this_rising))
            if len(this_falling) > 0:
                timestamps.extend(this_falling)
                labels.extend([f"{channel} OFF"] * len(this_falling))
        timestamps = np.asarray(timestamps)
        if len(labels) == 0:
            labels = np.asarray(labels, dtype="U1")
        else:
            labels = np.asarray(labels)
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        info = self.signals_info_dict[0, "nidq"]  # There are no events that are not in the nidq stream
        event_times = event_timestamps.astype(dtype) / float(info["sampling_rate"])
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        return None


def scan_files(dirname):
    """
    Scan for pairs of `.bin` and `.meta` files and return information about it.

    After exploring the folder, the segment index (`seg_index`) is construct as follow:
      * if only one `gate_num=0` then `trigger_num` = `seg_index`
      * if only one `trigger_num=0` then `gate_num` = `seg_index`
      * if both are increasing then seg_index increased by gate_num, trigger_num order.
    """
    info_list = []

    for root, dirs, files in os.walk(dirname):
        for file in files:
            if not file.endswith(".meta"):
                continue
            meta_filename = Path(root) / file
            bin_filename = meta_filename.with_suffix(".bin")
            if meta_filename.exists() and bin_filename.exists():
                meta = read_meta_file(meta_filename)
                info = extract_stream_info(meta_filename, meta)

                info["meta_file"] = str(meta_filename)
                info["bin_file"] = str(bin_filename)
                info_list.append(info)

    if len(info_list) == 0:
        raise FileNotFoundError(f"No appropriate combination of .meta and .bin files were detected in {dirname}")

    # the segment index will depend on both 'gate_num' and 'trigger_num'
    # so we order by 'gate_num' then 'trigger_num'
    # None is before any int
    def make_key(info):
        k0 = info["gate_num"]
        if k0 is None:
            k0 = -1
        k1 = info["trigger_num"]
        if k1 is None:
            k1 = -1
        return (k0, k1)

    order_key = list({make_key(info) for info in info_list})
    order_key = sorted(order_key)
    for info in info_list:
        info["seg_index"] = order_key.index(make_key(info))

    return info_list


def parse_spikeglx_fname(fname):
    """
    Parse recording identifiers from a SpikeGLX style filename.

    spikeglx naming follow this rules:
    https://github.com/billkarsh/SpikeGLX/blob/15ec8898e17829f9f08c226bf04f46281f106e5f/Markdown/UserManual.md#gates-and-triggers

    Example file name structure:
    Consider the filenames: `Noise4Sam_g0_t0.nidq.bin` or `Noise4Sam_g0_t0.imec0.lf.bin`
    The filenames consist of 3 or 4 parts separated by `.`
    1. "Noise4Sam_g0_t0" will be the `name` variable. This choosen by the user at recording time.
    2. "g0" is the "gate_num"
    3. "t0" is the "trigger_num"
    4. "nidq" or "imec0" will give the `device`
    5. "lf" or "ap" will be the `stream_kind`
       `stream_name` variable is the concatenation of `device.stream_kind`

    If CatGT is used, then the trigger numbers in the file names ("t0"/"t1"/etc.)
    will be renamed to "tcat". In this case, the parsed "trigger_num" will be set to "cat".

    This function is copied/modified from Graham Findlay.

    Notes:
        * Sometimes the original file name is modified by the user and "_g0_" or "_t0_"
          are manually removed. In that case gate_name and trigger_num will be None.

    Parameters
    ---------
    fname: str
        The filename to parse without the extension, e.g. "my-run-name_g0_t1.imec2.lf"

    Returns
    -------
    run_name: str
        The run name, e.g. "my-run-name".
    gate_num: int or None
        The gate identifier, e.g. 0.
    trigger_num: int | str or None
        The trigger identifier, e.g. 1. If CatGT is used, then the trigger_num will be set to "cat".
    device: str
        The probe identifier, e.g. "imec2"
    stream_kind: str or None
        The data type identifier, "lf" or "ap" or None
    """
    re_standard = re.findall(r"(\S*)_g(\d*)_t(\d*)\.(\S*).(ap|lf)", fname)
    re_tcat = re.findall(r"(\S*)_g(\d*)_tcat.(\S*).(ap|lf)", fname)
    re_nidq = re.findall(r"(\S*)_g(\d*)_t(\d*)\.(\S*)", fname)
    if len(re_standard) == 1:
        # standard case with probe
        run_name, gate_num, trigger_num, device, stream_kind = re_standard[0]
    elif len(re_tcat) == 1:
        # tcat case
        run_name, gate_num, device, stream_kind = re_tcat[0]
        trigger_num = "cat"
    elif len(re_nidq) == 1:
        # case for nidaq
        run_name, gate_num, trigger_num, device = re_nidq[0]
        stream_kind = None
    else:
        # the naming do not correspond lets try something more easy
        # example: sglx_xxx.imec0.ap
        re_else = re.findall(r"(\S*)\.(\S*).(ap|lf)", fname)
        re_else_nidq = re.findall(r"(\S*)\.(\S*)", fname)
        if len(re_else) == 1:
            run_name, device, stream_kind = re_else_nidq[0]
            gate_num, trigger_num = None, None
        elif len(re_else_nidq) == 1:
            # easy case for nidaq, example: sglx_xxx.nidq
            run_name, device = re_else_nidq[0]
            gate_num, trigger_num, stream_kind = None, None, None
        else:
            raise ValueError(f"Cannot parse filename {fname}")

    if gate_num is not None:
        gate_num = int(gate_num)
    if trigger_num is not None and trigger_num != "cat":
        trigger_num = int(trigger_num)

    return (run_name, gate_num, trigger_num, device, stream_kind)


def read_meta_file(meta_file):
    """parse the meta file"""
    with open(meta_file, mode="r") as f:
        lines = f.read().splitlines()

    meta = {}
    # Fix taken from: https://github.com/SpikeInterface/probeinterface/blob/
    # 19d6518fbc67daca71aba5e99d8aa0d445b75eb7/probeinterface/io.py#L649-L662
    for line in lines:
        split_lines = line.split("=")
        if len(split_lines) != 2:
            continue
        k, v = split_lines
        if k.startswith("~"):
            # replace by the list
            k = k[1:]
            v = v[1:-1].split(")(")[1:]
        meta[k] = v

    return meta


def extract_stream_info(meta_file, meta):
    """Extract info from the meta dict"""

    num_chan = int(meta["nSavedChans"])
    if "snsApLfSy" in meta:
        # AP and LF meta have this field
        ap, lf, sy = [int(s) for s in meta["snsApLfSy"].split(",")]
        has_sync_trace = sy == 1
    else:
        # NIDQ case
        has_sync_trace = False
    fname = Path(meta_file).stem
    run_name, gate_num, trigger_num, device, stream_kind = parse_spikeglx_fname(fname)

    if "imec" in fname.split(".")[-2]:
        device = fname.split(".")[-2]
        stream_kind = fname.split(".")[-1]
        stream_name = device + "." + stream_kind
        units = "uV"
        # please note the 1e6 in gain for this uV

        # metad['imroTbl'] contain two gain per channel  AP and LF
        # except for the last fake channel
        per_channel_gain = np.ones(num_chan, dtype="float64")
        if (
            "imDatPrb_type" not in meta
            or meta["imDatPrb_type"] == "0"
            or meta["imDatPrb_type"] in ("1015", "1016", "1022", "1030", "1031", "1032", "1100", "1121", "1300")
        ):
            # This work with NP 1.0 case with different metadata versions
            # https://github.com/billkarsh/SpikeGLX/blob/15ec8898e17829f9f08c226bf04f46281f106e5f/Markdown/Metadata_30.md
            if stream_kind == "ap":
                index_imroTbl = 3
            elif stream_kind == "lf":
                index_imroTbl = 4
            for c in range(num_chan - 1):
                v = meta["imroTbl"][c].split(" ")[index_imroTbl]
                per_channel_gain[c] = 1.0 / float(v)
            gain_factor = float(meta["imAiRangeMax"]) / 512
            channel_gains = gain_factor * per_channel_gain * 1e6
        elif meta["imDatPrb_type"] in ("21", "24", "2003", "2004", "2013", "2014"):
            # This work with NP 2.0 case with different metadata versions
            # https://github.com/billkarsh/SpikeGLX/blob/15ec8898e17829f9f08c226bf04f46281f106e5f/Markdown/Metadata_30.md#imec
            # We allow also LF streams for NP2.0 because CatGT can produce them
            # See: https://github.com/SpikeInterface/spikeinterface/issues/1949
            if "imChan0apGain" in meta:
                per_channel_gain[:-1] = 1 / float(meta["imChan0apGain"])
            else:
                per_channel_gain[:-1] = 1 / 80.0
            max_int = int(meta["imMaxInt"]) if "imMaxInt" in meta else 8192
            gain_factor = float(meta["imAiRangeMax"]) / max_int
            channel_gains = gain_factor * per_channel_gain * 1e6
        else:
            raise NotImplementedError("This meta file version of spikeglx" " is not implemented")
    else:
        device = fname.split(".")[-1]
        stream_kind = ""
        stream_name = device
        units = "V"
        channel_gains = np.ones(num_chan)

        # there are differents kinds of channels with different gain values
        mn, ma, xa, dw = [int(e) for e in meta["snsMnMaXaDw"].split(sep=",")]
        per_channel_gain = np.ones(num_chan, dtype="float64")
        per_channel_gain[0:mn] = 1.0 / float(meta["niMNGain"])
        per_channel_gain[mn : mn + ma] = 1.0 / float(meta["niMAGain"])
        # this scaling come from the code in this zip
        # https://billkarsh.github.io/SpikeGLX/Support/SpikeGLX_Datafile_Tools.zip
        # in file readSGLX.py line76
        # this is equivalent of 2**15
        gain_factor = float(meta["niAiRangeMax"]) / 32768
        channel_gains = per_channel_gain * gain_factor

    info = {}
    info["fname"] = fname
    info["meta"] = meta
    for k in ("niSampRate", "imSampRate"):
        if k in meta:
            info["sampling_rate"] = float(meta[k])
    info["num_chan"] = num_chan

    info["sample_length"] = int(meta["fileSizeBytes"]) // 2 // num_chan
    info["gate_num"] = gate_num
    info["trigger_num"] = trigger_num
    info["device"] = device
    info["stream_kind"] = stream_kind
    info["stream_name"] = stream_name
    info["units"] = units
    info["channel_names"] = [txt.split(";")[0] for txt in meta["snsChanMap"]]
    info["channel_gains"] = channel_gains
    info["channel_offsets"] = np.zeros(info["num_chan"])
    info["has_sync_trace"] = has_sync_trace

    if "nidq" in device:
        info["digital_channels"] = []
        info["analog_channels"] = [channel for channel in info["channel_names"] if not channel.startswith("XD")]
        # Digital/event channels are encoded within the digital word, so that will need more handling
        if meta.get("niXDChans1", "") != "":
            nixd_chans1_items = meta["niXDChans1"].split(",")
            for item in nixd_chans1_items:
                if ":" in item:
                    start, end = map(int, item.split(":"))
                    info["digital_channels"].extend([f"XD{i}" for i in range(start, end + 1)])
                else:
                    info["digital_channels"].append(f"XD{int(item)}")
    return info



"""
Test of neo.io.asciiimageio
"""

import os
import unittest
import quantities as pq
from neo.io import AsciiImageIO
import numpy as np


class TestAsciiImageIO(unittest.TestCase):

    def test_read_txt(self):
        img = ""
        img_list = []
        for frame in range(20):
            img_list.append([])
            for y in range(50):
                img_list[frame].append([])
                for x in range(50):
                    img += str(x)
                    img += "\t"
                    img_list[frame][y].append(x)
        img_list = np.array(img_list)
        file_name = "txt_test_file.txt"
        file = open(file_name, mode="w")
        file.write(str(img))
        file.close()

        object = AsciiImageIO(
            file_name="txt_test_file.txt",
            nb_frame=20,
            nb_row=50,
            nb_column=50,
            units="V",
            sampling_rate=1 * pq.Hz,
            spatial_scale=1 * pq.micrometer,
        )
        block = object.read_block()
        self.assertEqual(len(block.segments), 1)
        self.assertEqual(len(block.segments[0].imagesequences), 1)
        self.assertEqual(block.segments[0].imagesequences[0].shape, (20, 50, 50))
        self.assertEqual(block.segments[0].imagesequences[0].any(), img_list.any())

        file.close()
        os.remove(file_name)


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.openephysbinaryrawio import OpenEphysBinaryRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestOpenEphysBinaryRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = OpenEphysBinaryRawIO
    entities_to_download = ["openephysbinary"]
    entities_to_test = [
        "openephysbinary/v0.5.3_two_neuropixels_stream",
        "openephysbinary/v0.4.4.1_with_video_tracking",
        "openephysbinary/v0.5.x_two_nodes",
        "openephysbinary/v0.6.x_neuropixels_multiexp_multistream",
        "openephysbinary/v0.6.x_neuropixels_with_sync",
    ]

    def test_sync(self):
        rawio_with_sync = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=True
        )
        rawio_with_sync.parse_header()
        stream_name = [s_name for s_name in rawio_with_sync.header["signal_streams"]["name"] if "AP" in s_name][0]
        stream_index = list(rawio_with_sync.header["signal_streams"]["name"]).index(stream_name)

        # AP stream has 385 channels
        chunk = rawio_with_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 385

        rawio_no_sync = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=False
        )
        rawio_no_sync.parse_header()

        # AP stream has 384 channels
        chunk = rawio_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 384

    def test_no_sync(self):
        # requesting sync channel when there is none raises an error
        with self.assertRaises(ValueError):
            rawio_no_sync = OpenEphysBinaryRawIO(
                self.get_local_path("openephysbinary/v0.6.x_neuropixels_multiexp_multistream"), load_sync_channel=True
            )
            rawio_no_sync.parse_header()


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.plexon2rawio import Plexon2RawIO


class Plexon2IO(Plexon2RawIO, BaseFromRaw):
    """
    Class for reading data from Plexon PL2 files

    The IO is based on the Plexon2RawIO, see comments for memory optimization
    in neo.rawio.plexon2rawio.Plexon2RawIO

    """

    def __init__(self, filename):
        Plexon2RawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Neo is a package for representing electrophysiology data in Python,
together with support for reading a wide range of neurophysiology file formats
"""

import importlib.metadata

# this need to be at the begining because some sub module will need the version
__version__ = importlib.metadata.version("neo")

import logging

logging_handler = logging.StreamHandler()

from neo.core import *
from neo.io import *



"""
Tests of neo.io.baseio
"""

import unittest

from neo.core import objectlist
from neo.io.baseio import BaseIO
from neo.core import NeoReadWriteError


class TestIOObjects(unittest.TestCase):
    def test__raise_error_when_not_readable_or_writable(self):
        reader = BaseIO()
        for ob in objectlist:
            if ob not in BaseIO.readable_objects:
                meth = getattr(reader, "read_" + ob.__name__.lower())
                self.assertRaises(
                    NeoReadWriteError,
                    meth,
                )

            if ob not in BaseIO.writeable_objects:
                meth = getattr(reader, "write_" + ob.__name__.lower())
                self.assertRaises(NeoReadWriteError, meth, ())


if __name__ == "__main__":
    unittest.main()
